/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: BSearch.h$
 * $Date: Sun Aug 10 06:27:57 1997$
 *
 Description:
	Binary search lookup table.

	Optimized heavily for Lookup(). This is at the expense of performance
	is other methods. This class is intented to be used to access keyed
	data that can all be inserted at a configuration stage. The Compile()
	method to used to optimized all inserted data to an ordered linear
	array.  Subsequently, only lookup operations are performed using
	a binary search until the object is destroyed.
\*____________________________________________________________________________*/
//$HeaderTop:$

#ifndef BSEARCH_H_
#define BSEARCH_H_

#include <assert.h>

#include "PIString.h"
#include "DblList.h"

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
template<class T>
class BSearch
{
private:
	/* --- internal class for raw entries --- */
	class Entry
	{
	public:
		PIString sKey;
		T tElement;
		Entry( const char *pKey, T tTheElement )
		:	sKey( pKey ),
			tElement( tTheElement )
		{};
	};

	/* --- class data --- */
	DblList lRawEntries;
	int iCompiled;
	int iNumberOfEntries;
	Entry **ppLinear;
	
	/* --- internal functions ---- */
	static int fnSortEntryComparision( const void *pV1, const void *pV2 );
	static int fnSearchEntryComparision( const void *pV1, const void *pV2 );

	/* --- forbid copy constructor --- */
	BSearch( const BSearch & )
		: iCompiled( 0 ), iNumberOfEntries( 0 ), ppLinear( 0 )
		{ assert( 0 ); };	

public:
	BSearch();
	~BSearch();
	int Add( const char *pKey, T tElement );
	int Compile();
	inline T Lookup( const PIString &sKey ) const;
};

#include "BSearch.hpp"

#endif	/* BSEARCH_H_ */

