/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: Pi3Win.c$
 * $Date: Sun Aug 10 06:40:13 1997$
 *
 Description:

\*____________________________________________________________________________*/
/* $SourceTop:$ */

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <assert.h>

#include "resrc1.h"
#include "Dialogs.h"
#include "Config.h"

/*____________________________________________________________________________*\
 *
 Description:
	Definitions and global values
\*____________________________________________________________________________*/
static HWND hWndThreads = 0;
static HWND hWndInstallDir = 0;
static HWND hWndRecv = 0;
static HWND hWndSend = 0;
static HWND hWndKeepOpen = 0;
static HWND hWndReverse = 0;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
	LPARAM lParam)
{
	switch (uMsg)
		{
		case WM_INITDIALOG:
			{
			const char *pT = AD_lookupValue( __pConfig, "General", "Threads" );
			const char *pI = AD_lookupValue( __pConfig, "General", "InstallDir" );
			const char *pR = AD_lookupValue( __pConfig, "General", "RecvTimeout" );
			const char *pS = AD_lookupValue( __pConfig, "General", "SendTimeout" );
			const char *pK = AD_lookupValue( __pConfig, "General", "KeepOpen" );
			const char *pD = AD_lookupValue( __pConfig, "General", 
				"DNSReverseLookup" );
			hWndThreads = GetDlgItem( hDlg, IDC_THREADS );
			hWndInstallDir = GetDlgItem( hDlg, IDC_INSTALLDIR );
			hWndRecv = GetDlgItem( hDlg, IDC_RECV );
			hWndSend = GetDlgItem( hDlg, IDC_SEND );
			hWndKeepOpen = GetDlgItem( hDlg, IDC_KEEPALIVE );
			hWndReverse = GetDlgItem( hDlg, IDC_REVERSE );
			assert( hWndThreads );
			assert( hWndInstallDir && hWndRecv && hWndSend );
			assert( hWndKeepOpen && hWndReverse );

			assert( pT && pI && pR && pS && pK && pD );

			::SendMessage( hWndThreads, WM_SETTEXT, 0, (LPARAM)pT );
			::SendMessage( hWndInstallDir, WM_SETTEXT, 0, (LPARAM)pI );
			::SendMessage( hWndRecv, WM_SETTEXT, 0, (LPARAM)pR );
			::SendMessage( hWndSend, WM_SETTEXT, 0, (LPARAM)pS );
			if ( !stricmp( pK, "On" ) )
				{ ::SendMessage( hWndKeepOpen, BM_SETCHECK, BST_CHECKED, 0 ); }
			else
				{ ::SendMessage( hWndKeepOpen, BM_SETCHECK, BST_UNCHECKED, 0 ); };
			if ( *pD )
				{ ::SendMessage( hWndReverse, BM_SETCHECK, BST_CHECKED, 0 ); }
			else
				{ ::SendMessage( hWndReverse, BM_SETCHECK, BST_UNCHECKED, 0 ); };
			}
			break;

		case WM_COMMAND:
			if ( GET_WM_COMMAND_CMD(wParam, lParam)==EN_CHANGE 
					|| GET_WM_COMMAND_CMD(wParam, lParam)==BN_CLICKED )
				{ 
				PropSheet_Changed( GetParent( hDlg ), hDlg );
				};
			break;

		case WM_NOTIFY:
			switch( ((NMHDR *)lParam)->code )
				{
				case PSN_QUERYCANCEL:
					return AD_close( hDlg );
				
				case PSN_KILLACTIVE:
					{
					enum { BUF_SIZE=1023 };
					char szBuf[BUF_SIZE+1];

					/*
					** Threads
					*/
					SendMessage( hWndThreads, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					int iNumThreads = atoi( szBuf );
					if ( iNumThreads<1 || iNumThreads>512 )
						{
						/*
						** Give message and abort change
						*/
						MessageBox( hDlg, "Threads must be in the range [1..512]",
							"Threads Field Error", MB_OK );
						SetFocus( hWndThreads );
						SendMessage( hWndThreads, EM_SETSEL, 0, -1 );
						SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
						return (TRUE);
						};
					AD_replaceValue( __pConfig, "General", "Threads", szBuf ); 

					::SendMessage( hWndInstallDir, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					AD_replaceValue( __pConfig, "General", "InstallDir", szBuf ); 
					::SendMessage( hWndRecv, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					AD_replaceValue( __pConfig, "General", "RecvTimeout", szBuf ); 
					::SendMessage( hWndSend, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					AD_replaceValue( __pConfig, "General", "SendTimeout", szBuf ); 
					if ( ::SendMessage( hWndKeepOpen, BM_GETCHECK, 0, 0 )==BST_CHECKED )
						{ AD_replaceValue( __pConfig, "General", "KeepOpen", "On" ); }
					else
						{ AD_replaceValue( __pConfig, "General", "KeepOpen", "Off" ); };
					if ( ::SendMessage( hWndReverse, BM_GETCHECK, 0, 0 )==BST_CHECKED )
						{ AD_replaceValue( __pConfig, "General", "DNSReverseLookup", "Yo" ); }
					else
						{ AD_replaceValue( __pConfig, "General", "DNSReverseLookup", "" ); };
					};
					SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
					return (TRUE);

				case PSN_APPLY:
					if ( !AD_apply( hDlg ) )
						{
						SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID );
						return (TRUE);
						};
					
					if( ((PSHNOTIFY *)lParam)->lParam == TRUE )
						{
						return AD_close( hDlg );
						};
					break;

				};
			break;

		default:
				return (FALSE);
		};

	return (TRUE);
}

/*____________________________________________________________________________*\
 *
 Description:
	Definitions and global values
\*____________________________________________________________________________*/
DLGPROC fnPerformance = (DLGPROC)DialogProc;
