/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: Pi3Win.c$
 * $Date: Sun Aug 10 06:40:13 1997$
 *
 Description:

\*____________________________________________________________________________*/
/* $SourceTop:$ */

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <assert.h>
#include <stdio.h>

#include "DblList.h"
#include "PIString.h"

#include "resrc1.h"
#include "Dialogs.h"
#include "Config.h"

/*____________________________________________________________________________*\
 *
 Description:
	Definitions and global values
\*____________________________________________________________________________*/
static HWND hWndRealm = 0;
static HWND hWndList = 0;
static HWND hWndUser = 0;
static HWND hWndPassword = 0;
static HWND hWndConfirm = 0;
static HWND hWndAdd = 0;
static HWND hWndReplace = 0;
static HWND hWndDelete = 0;
static DblList lUsers;
static DblList lPasswords;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Not very efficient, but it doesn't really need to be.
\*____________________________________________________________________________*/
const char __pBase64[65]="\
ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static void Internal_base64Encode( const char *pPlain, PIString &sResult )
{
	assert( pPlain );
	int iLen = strlen( pPlain );
	enum { START, TWO, FOUR, SIX } eState = START;
	int i=0;
	for( ;; )
		{
		int iPlain;
		int iBase64;
		char c = (i<iLen) ? pPlain[i] : 0 ;
		switch( eState )
			{
			case START:
				if ( i>=iLen ) { return; };
				iBase64 = (c >> 2);
				iPlain = (c << 4 ) & 0x30;
				i++;
				eState = TWO;
				break;

			case TWO:
				iBase64 = iPlain | ( c >> 4 );
				iPlain = (c << 2 ) & 0x3C;
				i++;
				eState = FOUR;
				break;
				
			case FOUR:
				iBase64 = iPlain | ( c >> 6 );
				iPlain = c & 0x3F;
				i++;
				eState = SIX;
				break;
				
			case SIX:
				iBase64 = iPlain;
				eState = START; 
				break;

			default:;
			};
		
		sResult.Concatenate( __pBase64[iBase64 & 0x3F] );

		if ( i>iLen )
			{
			switch( eState )
				{
				case START:
				case TWO:
					sResult.Concatenate( '=' );
				case FOUR:
					sResult.Concatenate( '=' );
				case SIX:
					sResult.Concatenate( '=' );
					return;
				default:;
					assert( 0 );
				};
			};
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Not very efficient, but it doesn't really need to be.
\*____________________________________________________________________________*/
const char __pPlain[128]={
/*   0.. 15 */ -1, -1, -1, -1, -1, -1, -1 ,-1, -1, -1, -1, -1, -1, -1, -1, -1,
/*  16.. 31 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
/*  32.. 47 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
/*  48.. 63 */ 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1,  0, -1, -1,
/*  64.. 79 */ -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
/*  80.. 95 */ 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
/*  96..111 */ -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
/* 112..127 */ 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1
};
static void Internal_base64Decode( const char *pBase64, PIString &sResult )
{
	assert( pBase64 );
	int iLen = strlen( pBase64 );
	enum { START, TWO, FOUR, SIX } eState = START;
	int iDone = 0;
	for( int i=0; i<iLen; i++)
		{
		char c = pBase64[i];
		if ( !isalnum(c) && (c!='+') && (c!='/') ) { break; };
		int iLast;
		int iPlain = __pPlain[c];
		assert( iPlain!= -1 );
		switch( eState )
			{
			case START:
				iLast = ( iPlain << 2 ) & 0xFC;
				eState = TWO;
				break;

			case TWO:
				sResult.Concatenate( iLast | ( (iPlain >> 4 ) & 0x03) );
				iLast = ( iPlain << 4 ) & 0xF0;
				eState = FOUR;
				break;
				
			case FOUR:
				sResult.Concatenate( iLast | ( (iPlain >> 2 ) & 0x0F) );
				iLast = ( iPlain << 6 ) & 0xC0;
				eState = SIX;
				break;
				
			case SIX:
				sResult.Concatenate( iLast | ( iPlain & 0x3F) );
				eState = START; 
				break;

			default:;
			};

		if ( c=='=')
			{ break; };
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_addLine( const char *pLine )
{
	assert( pLine );
	assert( lUsers.Size()==lPasswords.Size() );

	int i;
	for( i=0; pLine[i] && !isspace(pLine[i]); i++ );
	
	/* --- scan to tab --- */
	lUsers.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );

	/* --- scan over whitespace to start of Media type/subtype --- */
	for( ; pLine[i] && isspace(pLine[i]); i++ );
	lPasswords.Append( (DblList::type)PI_NEW( PIString( &(pLine[i]) ) ) );	

	assert( lUsers.Size()==lPasswords.Size() );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_clearArrays()
{
	DblListIterator x( lUsers );
	DblListIterator y( lPasswords );
	for( ; !x.BadIndex(); x++, y++ )
		{
		assert( !x.BadIndex() && !y.BadIndex() );
		PI_DELETE( (PIString *)x.Current() );
		PI_DELETE( (PIString *)y.Current() );
		};
	lUsers.Clear();
	lPasswords.Clear();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_reloadArrays()
{
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];

	Internal_clearArrays();

	/*
	** Get number of items
	*/
	int iCount = SendMessage( hWndList, LB_GETCOUNT, 0, 0 );

	/* 
	** Insert from listbox into internal arrays
	*/
	int i;
	for( i=0; i<iCount; i++ )
		{
		SendMessage( hWndList, LB_GETTEXT, i, (LPARAM)szBuf );
		Internal_addLine( szBuf );
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_updateButtons( WPARAM wParam, LPARAM lParam, HWND hWndEdit1, 
	HWND hWndEdit2, const char *pUser )
{
	if ( GET_WM_COMMAND_CMD(wParam, lParam)!=EN_CHANGE )
		{ return; };

	/*
	** If the file extension changed then check if this is an existing
	** file extension
	*/
	int iExistingIndex = -1;
	assert( pUser );
	int iIndex = 0;
	if ( *pUser )
		{
		for( DblListIterator i( lUsers ); !i.BadIndex(); i++, iIndex++ )
			{
			PIString *pString = (PIString *)i.Current();
			assert( pString );
			if ( (*pString)==pUser )
				{
				iExistingIndex = iIndex;
				break;
				};
			};
		};

	/*
	** Remove listbox selection or scroll to correct line
	** as appropriate 
	*/
	SendMessage( hWndList, LB_SETCURSEL, iExistingIndex, 0 );

	/*
	** Set buttons
	*/
	if (	SendMessage( hWndEdit1, EM_LINELENGTH, 0, 0 )==0 ||
			SendMessage( hWndEdit2, EM_LINELENGTH, 0, 0 )==0 ||
			SendMessage( hWndConfirm, EM_LINELENGTH, 0, 0 )==0 )
		{
		if ( IsWindowEnabled( hWndAdd ) ) EnableWindow( hWndAdd, FALSE );
		if ( IsWindowEnabled( hWndReplace ) ) EnableWindow( hWndReplace, FALSE );
		}
	else
		{
		if ( iExistingIndex>-1 )	/* --- disable add, enable replace --- */
			{
			if ( IsWindowEnabled( hWndAdd ) ) EnableWindow( hWndAdd, FALSE );
			if ( !IsWindowEnabled( hWndReplace ) ) EnableWindow( hWndReplace, TRUE );
			}
		else						/* --- disable replace, enable add --- */
			{
			if ( !IsWindowEnabled( hWndAdd ) ) EnableWindow( hWndAdd, TRUE );
			if ( IsWindowEnabled( hWndReplace ) ) EnableWindow( hWndReplace, FALSE );
			};
		};

	/*
	** If there is a selected item in the listbox then
	** enable delete, otherwise disable delete	
	*/
	if ( SendMessage( hWndList, LB_GETCURSEL, 0, 0 )==LB_ERR )
		{ EnableWindow( hWndDelete, FALSE ); }
	else
		{ EnableWindow( hWndDelete, TRUE ); };
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_loadItems( const char *pRealm )
{
	assert( pRealm );

	/*
	** Clear any existing items out of listbox
	*/
	SendMessage( hWndList, LB_RESETCONTENT, 0, 0 );

	/*
	** Load all users into list box
	*/
	int iTmp = 0;
	for(;; iTmp++ )
		{
		PIString sSection( "__" );
		sSection.Concatenate( pRealm );
		const char *pTmp = AD_lookupValueEx( __pConfig, sSection, "Base64", iTmp );
		if ( !pTmp )
			{ break; };

		PIString sPlain;
		Internal_base64Decode( pTmp, sPlain );
		char *pPlain = (char *)(const char *)sPlain;	
		int iLen = sPlain.Len();
		int i;
		for( i=0; i<iLen && pPlain[i]!=':'; i++);
		if ( pPlain[i]==':' ) { pPlain[i]='\t'; };

		SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)pPlain );
		};

	/*
	** Clear out edit fields
	*/
	::SendMessage( hWndUser, WM_SETTEXT, 0, (LPARAM)"" );
	::SendMessage( hWndPassword, WM_SETTEXT, 0, (LPARAM)"" );
	::SendMessage( hWndConfirm, WM_SETTEXT, 0, (LPARAM)"" );

	Internal_reloadArrays();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int Internal_canDeleteRealm( HWND hDlg, const char *pRealm )
{
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];
	int iTmp = 0;

	assert( pRealm );

	/*
	** Check if realm is used by users
	*/
	iTmp = 0;
	for(;; iTmp++ )
		{
		const char *pTmp = AD_lookupValueEx( __pConfig, "AccessControl", 
			"URLs", iTmp );

		if ( !pTmp ) { break; };

		int i;
		for( i=0; pTmp[i] && pTmp[i]!='\t'; i++ );

		PIString sRealm( pTmp, i );
		
		if ( sRealm==pRealm )
			{
			sprintf( szBuf, "Realm '%s' is being used for access control. \
Remove all Realm-URL Path mappings in 'Access Control' before removing \
this realm.", pRealm );
			MessageBox( hDlg, szBuf, "Remove Realm", MB_OK | MB_ICONEXCLAMATION );
			return 0;
			};
		};

	/*
	** Check if realm is used by a mapping
	*/
	iTmp = 0;
	for(;; iTmp++ )
		{
		const char *pTmp = AD_lookupValueEx( __pConfig, "Mappings", 
			"Mapping", iTmp );

		if ( !pTmp ) { break; };

		int i=0;
		for( ; pTmp[i] && isspace(pTmp[i]); i++ );	
		for( ; pTmp[i] && pTmp[i]!='\t'; i++ );
		for( ; pTmp[i] && isspace(pTmp[i]); i++ );
		for( ; pTmp[i] && pTmp[i]!='\t'; i++ );
		for( ; pTmp[i] && isspace(pTmp[i]); i++ );
		for( ; pTmp[i] && pTmp[i]!='\t'; i++ );
		for( ; pTmp[i] && isspace(pTmp[i]); i++ );	
		pTmp = &( pTmp[i] );
		for( ; pTmp[i] && pTmp[i]!='\t'; i++ );	/* read length of realm */

		PIString sRealm( pTmp, i );
		
		if ( sRealm==pRealm )
			{
			sprintf( szBuf, "Realm '%s' is being used for directory mapping. \
Remove all Realm-URL Path mappings in 'Mapping' before removing \
this realm.", pRealm );
			MessageBox( hDlg, szBuf, "Remove Realm", MB_OK | MB_ICONEXCLAMATION );
			return 0;
			};
		};

	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_saveItems( const char *pRealm )
{
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];

	assert( pRealm );

	int i;
	int iCount = SendMessage( hWndList, LB_GETCOUNT, 0, 0 );

	/*
	** Save all user types
	*/
	/* --- remove old ones first --- */
	PIString sSection( "__" );
	sSection.Concatenate( pRealm );
	AD_deleteValue( __pConfig, sSection, "Base64" );
	for( i=0; i<iCount; i++ )
		{
		*szBuf = '\0';
		szBuf[BUF_SIZE] = '\0';
		SendMessage( hWndList, LB_GETTEXT, i, (LPARAM)szBuf );

		int iLen = strlen( szBuf );
		int j;
		for( j=0; j<iLen && szBuf[j]!='\t'; j++ );
		if ( szBuf[j]=='\t' ) { szBuf[j] = ':'; };
		PIString sBase64;
		Internal_base64Encode( szBuf, sBase64 );
		
		AD_addValue( __pConfig, sSection, "Base64", sBase64 );
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static HWND __hWndDlgOK = 0;
static HWND __hWndDlgRealm = 0;
static BOOL CALLBACK NewRealmDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
	LPARAM lParam)
{
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];

	switch (uMsg)
		{
		case WM_INITDIALOG:
			__hWndDlgOK = GetDlgItem( hDlg, IDOK );
			__hWndDlgRealm = GetDlgItem( hDlg, IDC_REALM );
			assert( __hWndDlgOK && __hWndDlgRealm );
			SendMessage( __hWndDlgRealm, EM_SETLIMITTEXT, 256, 0 );
			EnableWindow( __hWndDlgOK, FALSE );
			break;

		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
				{
				case IDOK:
					{
					/*
					** Validate content of edit field
					*/
					*szBuf = '\0';
					SendMessage( __hWndDlgRealm, WM_GETTEXT, BUF_SIZE, 
						(LPARAM)szBuf );
					PIString sRealm( szBuf );
					int i;
					int iCount = SendMessage( hWndList, CB_GETCOUNT, 0, 0 );
					for( i=0; i<iCount; i++ )
						{
						*szBuf = '\0';
						SendMessage( hWndRealm, CB_GETLBTEXT, i, (LPARAM)szBuf );
						if ( sRealm == szBuf )
							{
							sprintf( szBuf, "Realm '%s' already exists.", (const char *)sRealm );
							MessageBox( hDlg, szBuf, "New Realm", MB_OK );
							return (TRUE);
							};
						};

					/*
					** Can't use '(none)'
					*/
					if ( !stricmp( sRealm, "(none)" ) )
						{
						sprintf( szBuf, "'%s' is a reserved realm.",
							(const char *)sRealm );
						::MessageBox( hDlg, szBuf,  "Remove Realm",
							MB_OK | MB_ICONEXCLAMATION );
						::SetFocus( __hWndDlgRealm );
						::SendMessage( __hWndDlgRealm, EM_SETSEL, 0, -1 );
						return (TRUE);
						};

					/*
					** Add new realm
					*/
					SendMessage( hWndRealm, CB_ADDSTRING, 0, (LPARAM)(const char *)sRealm );

					/*
					** Save old items
					*/
					*szBuf = '\0';
					SendMessage( hWndRealm, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					Internal_saveItems( szBuf );

					int iSel = SendMessage( hWndRealm, CB_FINDSTRINGEXACT, (WPARAM)-1, 
						(LPARAM)(const char *)sRealm );
					if ( iSel!=CB_ERR )
						{
						SendMessage( hWndRealm, CB_SETCURSEL, iSel, 0 ); 
						if ( SendMessage( hWndRealm, CB_GETCURSEL, 0, 0 )!=
							CB_ERR )
							{
							EnableWindow( hWndUser, TRUE );
							EnableWindow( hWndPassword, TRUE );
							EnableWindow( hWndConfirm, TRUE );
							};
						Internal_loadItems( sRealm );
						}		
					else
						{ assert( 0 ); };

					};
					EndDialog( hDlg, 1 );
					break;

				case IDCANCEL:
					EndDialog( hDlg, 0 );
					break;

				case IDC_REALM:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)==EN_CHANGE )
						{
						if ( SendMessage( GET_WM_COMMAND_HWND(wParam, lParam ), EM_LINELENGTH,
							0, 0 )>0 )
							{ EnableWindow( __hWndDlgOK, TRUE ); }
						else
							{ EnableWindow( __hWndDlgOK, FALSE ); }
						}
					else
						return (FALSE);
					break;

				default:;
				};

		default:
			return (FALSE);
		};

	return (TRUE);
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
	LPARAM lParam)
{
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];
	szBuf[BUF_SIZE]='\0';
	int aTabstop[2];

	switch (uMsg)
		{
		case WM_INITDIALOG:
			AD_hourglassOn();
			{
			hWndRealm = GetDlgItem( hDlg, IDC_REALM );
			hWndList = GetDlgItem( hDlg, IDC_LIST );
			hWndUser = GetDlgItem( hDlg, IDC_USER );
			hWndPassword = GetDlgItem( hDlg, IDC_PASSWORD );
			hWndConfirm = GetDlgItem( hDlg, IDC_CONFIRM );
			hWndAdd = GetDlgItem( hDlg, IDC_ADD );
			hWndReplace = GetDlgItem( hDlg, IDC_REPLACE );
			hWndDelete= GetDlgItem( hDlg, IDC_DELETE );

			assert( hWndRealm && hWndList && hWndUser && hWndPassword );
			assert( hWndConfirm && hWndAdd && hWndReplace && hWndDelete );

			EnableWindow( hWndAdd, FALSE );
			EnableWindow( hWndReplace, FALSE );
			EnableWindow( hWndDelete, FALSE );

			/*
			** Limit edit field lengths, really just a sanity thing
			*/
			SendMessage( hWndUser, EM_SETLIMITTEXT, 256, 0 );
			SendMessage( hWndPassword, EM_SETLIMITTEXT, 256, 0 );
			SendMessage( hWndConfirm, EM_SETLIMITTEXT, 256, 0 );

			/*
			** Setup tabstops
			*/
			aTabstop[0] = 0;
			aTabstop[1] = 750;
			SendMessage( hWndList, LB_SETTABSTOPS, 2, (LPARAM)&aTabstop );

			/*
			** Load all users into realm box
			*/
			int iTmp = 0;
			for(;; iTmp++ )
				{
				const char *pTmp = AD_lookupValueEx( __pConfig, "Realms", "Basic", 
					iTmp );
				if ( !pTmp ) { break; };

				SendMessage( hWndRealm, CB_ADDSTRING, 0, (LPARAM)pTmp );
				};

			if ( SendMessage( hWndRealm, CB_GETCOUNT, 0, 0 )>0 )
				{
				SendMessage( hWndRealm, CB_SETCURSEL, 0, 0 );
				*szBuf = '\0';
				SendMessage( hWndRealm, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );

				/*
				** Load items into list box
				*/
				Internal_loadItems( szBuf );
				}
			else
				{
				/* --- no realms, disable edit fields --- */
				EnableWindow( hWndUser, FALSE );
				EnableWindow( hWndPassword, FALSE );
				EnableWindow( hWndConfirm, FALSE );
				};
			}
			AD_hourglassOff();
			break;

		case WM_DESTROY:
			{ Internal_clearArrays(); };
			break;

		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
				{
				case IDC_REALM:
					switch( GET_WM_COMMAND_CMD(wParam, lParam) )
						{
						case CBN_DROPDOWN:
							/*
							** Save current realm users
							*/
							*szBuf = '\0';
							SendMessage( hWndRealm, WM_GETTEXT, BUF_SIZE, 
								(LPARAM)szBuf );
							Internal_saveItems( szBuf );
							break;

						case CBN_SELCHANGE:
							/*
							** Load new realm users
							*/
							*szBuf = '\0';
							SendMessage( hWndRealm, WM_GETTEXT, 
								BUF_SIZE, (LPARAM)szBuf );
							if ( SendMessage( hWndRealm, CB_GETCURSEL, 0, 0 )!=
								CB_ERR )
								{
								EnableWindow( hWndUser, TRUE );
								EnableWindow( hWndPassword, TRUE );
								EnableWindow( hWndConfirm, TRUE );
								};
							Internal_loadItems( szBuf );
							break;

						default:;
						};
					break;

				case IDC_NEW:
					if ( DialogBox(
						NULL,
						MAKEINTRESOURCE( IDD_NEWREALM ),
						hDlg,
						(DLGPROC)NewRealmDialogProc
						) )
						{ PropSheet_Changed( GetParent( hDlg ), hDlg ); };
					break;


				case IDC_REMOVE:
					{
					int iSel = SendMessage( hWndRealm, CB_GETCURSEL, 0, 0 );
					if ( iSel==CB_ERR )
						{ break; };

					SendMessage( hWndRealm, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
		
					if ( !Internal_canDeleteRealm( hDlg, szBuf ) )
						{ break; };

					SendMessage( hWndRealm, CB_DELETESTRING, iSel, 0 );
					PIString sSection( "__" );
					sSection.Concatenate( szBuf );
					AD_deleteValue( __pConfig, sSection, "Base64" );
					Internal_clearArrays();
					SendMessage( hWndList, LB_RESETCONTENT, 0, 0 );
					SendMessage( hWndUser, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndPassword, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndConfirm, WM_SETTEXT, 0, (LPARAM)"" );
					PropSheet_Changed( GetParent( hDlg ), hDlg );
					if ( SendMessage( hWndRealm, CB_GETCURSEL, 0, 0 )==CB_ERR )
						{
						SendMessage( hWndRealm, CB_SETCURSEL, (WPARAM)iSel, 0 );
						EnableWindow( hWndUser, FALSE );
						EnableWindow( hWndPassword, FALSE );
						EnableWindow( hWndConfirm, FALSE );
						};
					};
					break;

				case IDC_ADD:
					{
					int i;
					*szBuf = '\0';
					SendMessage( hWndUser, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					/*
					** Username cannot contain ':' character
					*/
					for( i=0; i<BUF_SIZE && szBuf[i] && szBuf[i]!=':'; i++ );
					if ( szBuf[i]==':' ) 
						{
						MessageBox( hDlg, "A username may not contain the colon (:) \
character.", "Add User", 
							MB_OK | MB_ICONEXCLAMATION );
						SetFocus( hWndUser );
						SendMessage( hWndUser, EM_SETSEL, 0, -1 );
						break;
						};
					for( i=0; i<BUF_SIZE && szBuf[i] && !isspace( szBuf[i] ); i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					PIString sUser( szBuf );
					szBuf[i++] = '\t';
					SendMessage( hWndPassword, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					PIString sPassword( &(szBuf[i]) );
					SendMessage( hWndConfirm, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					PIString sConfirm( &(szBuf[i]) );
					if ( !(sPassword==sConfirm) )
						{
						MessageBox( hDlg, "Passwords do not match", "New User", 
							MB_OK | MB_ICONEXCLAMATION );
						SetFocus( hWndPassword );
						SendMessage( hWndPassword, EM_SETSEL, 0, -1 );
						break;
						};
					lUsers.Append( (DblList::type)PI_NEW( PIString( sUser ) ) );
					lPasswords.Append( (DblList::type)PI_NEW( PIString ( sPassword ) ) );
					SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)szBuf );

					/*
					** Clear edit fields, focus to list
					*/
					SendMessage( hWndUser, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndPassword, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndConfirm, WM_SETTEXT, 0, (LPARAM)"" );
					Internal_reloadArrays();
					int iIndex = SendMessage( hWndList, LB_FINDSTRINGEXACT, 0, (LPARAM)szBuf );
					if ( iIndex!=LB_ERR ) SendMessage( hWndList, LB_SETTOPINDEX, iIndex, 0 );
					SetFocus( hWndList );
					PropSheet_Changed( GetParent( hDlg ), hDlg );
					};
					break;
	
				case IDC_REPLACE:
					{
					/*
					** Get selected item index 
					*/
					int iCurSel = SendMessage( hWndList, LB_GETCURSEL, 0, 0 );

					/*
					** Get text to add and add it to internal lists only
					*/
					int i;
					*szBuf = '\0';
					SendMessage( hWndUser, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					/*
					** Username cannot contain ':' character
					*/
					for( i=0; i<BUF_SIZE && szBuf[i] && szBuf[i]!=':'; i++ );
					if ( szBuf[i]==':' ) 
						{
						MessageBox( hDlg, "A username may not contain the colon (:) \
character.", "Add User", 
							MB_OK | MB_ICONEXCLAMATION );
						SetFocus( hWndUser );
						SendMessage( hWndUser, EM_SETSEL, 0, -1 );
						break;
						};
					for( i=0; i<BUF_SIZE && szBuf[i] && !isspace( szBuf[i] ); i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					PIString sUser( szBuf );
					szBuf[i++] = '\t';
					SendMessage( hWndPassword, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					PIString sPassword( &(szBuf[i]) );
					SendMessage( hWndConfirm, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					PIString sConfirm( &(szBuf[i]) );
					if ( !(sPassword==sConfirm) )
						{
						MessageBox( hDlg, "Passwords do not match", "New User", 
							MB_OK | MB_ICONEXCLAMATION );
						SetFocus( hWndPassword );
						SendMessage( hWndPassword, EM_SETSEL, 0, -1 );
						break;
						};
					lUsers.Append( (DblList::type)PI_NEW( PIString( sUser ) ) );
					lPasswords.Append( (DblList::type)PI_NEW( PIString ( sPassword ) ) );

					/*
					** Delete the item from the list and internal lists
					*/
					SendMessage( hWndList, LB_DELETESTRING,	iCurSel, 0 );
					PI_DELETE( (PIString *)lUsers.Detach( iCurSel) );
					PI_DELETE( (PIString *)lPasswords.Detach( iCurSel) );
					
					/*
					** Add new item to listbox
					*/
					SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)szBuf );

					/*
					** Clear edit fields, focus to list
					*/
					SendMessage( hWndUser, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndPassword, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndConfirm, WM_SETTEXT, 0, (LPARAM)"" );
					Internal_reloadArrays();
					int iIndex = SendMessage( hWndList, LB_FINDSTRINGEXACT, 0, (LPARAM)szBuf );
					if ( iIndex!=LB_ERR ) SendMessage( hWndList, LB_SETTOPINDEX, iIndex, 0 );
					SetFocus( hWndList );
					PropSheet_Changed( GetParent( hDlg ), hDlg );
					};
					break;

				case IDC_DELETE:
					{
					int iCurSel = SendMessage( hWndList, LB_GETCURSEL, 0, 0 );
					SendMessage( hWndList, LB_DELETESTRING,	iCurSel, 0 );
					PI_DELETE( (PIString *)lUsers.Detach( iCurSel) );
					PI_DELETE( (PIString *)lPasswords.Detach( iCurSel) );
			
					/*
					** Clear edit fields, focus to list
					*/
					SendMessage( hWndUser, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndPassword, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndConfirm, WM_SETTEXT, 0, (LPARAM)"" );
					SetFocus( hWndList );
					PropSheet_Changed( GetParent( hDlg ), hDlg );
					};
					break;

				case IDC_USER:
					*szBuf = '\0';
					SendMessage( hWndUser, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf ); 
					Internal_updateButtons( wParam, lParam, hWndUser, 
						hWndPassword, szBuf );
					break;

				case IDC_PASSWORD:
				case IDC_CONFIRM:
					*szBuf = '\0';
					SendMessage( hWndUser, WM_GETTEXT, BUF_SIZE,(LPARAM)szBuf );
					Internal_updateButtons( wParam, lParam, hWndPassword, 
						hWndUser, szBuf );
					break;

				case IDC_LIST:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)==LBN_SELCHANGE )
						{
						int iSel = SendMessage( hWndList, LB_GETCURSEL, 0, 0 );
						if ( iSel!=LB_ERR )
							{
							assert( iSel>=0 && iSel<lUsers.Size() );
							EnableWindow( hWndDelete, TRUE );
							SendMessage( hWndUser, WM_SETTEXT, 0,
								(LPARAM)(const char *)*((PIString *)lUsers[iSel]) );
							SendMessage( hWndPassword, WM_SETTEXT, 0,
								(LPARAM)(const char *)*((PIString *)lPasswords[iSel]) );
							SendMessage( hWndConfirm, WM_SETTEXT, 0,
								(LPARAM)(const char *)*((PIString *)lPasswords[iSel]) );
							};
						};
					break;

				default:;		
				}
			break;

		case WM_NOTIFY:
			{
			switch( ((NMHDR *)lParam)->code )
				{
				case PSN_QUERYCANCEL:
					return AD_close( hDlg );

				case PSN_KILLACTIVE:
					/*
					** Save realms
					*/
					{
					AD_deleteValue( __pConfig, "Realms", "Basic" );
					int i;
					int iCount = SendMessage( hWndRealm, CB_GETCOUNT, 0, 0 );
					for( i=0; i<iCount; i++ )
						{
						*szBuf = '\0';
						SendMessage( hWndRealm, CB_GETLBTEXT, i, (LPARAM)szBuf );
						AD_addValue( __pConfig, "Realms", "Basic", szBuf );
						};
					};

					/*
					** Save values
					*/
					*szBuf = '\0';
					SendMessage( hWndRealm, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					Internal_saveItems( szBuf );
					SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
					break;

					default:;
					};
				};
				break;

		default:
				return (FALSE);
		};

	return (TRUE);
}

/*____________________________________________________________________________*\
 *
 Description:
	Definitions and global values
\*____________________________________________________________________________*/
DLGPROC fnUsers = (DLGPROC)DialogProc;
