/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: BscProxy.cpp$
 * $Date: Sun Aug 10 06:36:43 1997$
 *
 Description:
	Simple HTTP proxying.

    The caching function is still under development.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <ctype.h>

#include "DblList.h"
#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name BasicProxyClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor BasicProxy_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name BasicProxy
		Class BasicProxyClass
	</Object>

/*___+++CNF_END+++___*/
#endif
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	BasicProxy

Description:
	Perform simple proxy requests. This handler allows some filtering of
	forward client headers, but no caching. Proxy requests use HTTP/1.0

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IOObject
<TD>-
<TD>&ltA Pi3 Object&gt
<TD>Prototype IO object for proxy forwarding
<TD>IOObject "TCPIPClient"

<TR>
<TD>CacheRoot
<TD>-
<TD>&ltdirectory path&gt
<TD>Relative or Absolute path to cache root
<TD>CacheRoot "Cache"

<!--

AllowHosts is not documented for now because its still under development

<TR>
<TD>AllowHosts
<TD>-
<TD>&ltdirectory path&gt
<TD>Relative or Absolute path to cache root
<TD>CacheRoot "Cache"

-->

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>IOObject</H5>
Specifies the protocol IO object used to connect to remote HTTP servers. This
object can be a protocol IO implementor (such as TCP/IP or named pipes) or an
IO filter which manipulates the data en route to the server.

<H5>CacheRoot</H5>
Specifies the path to the directory where cached data will be kept for
proxy caching.
If this value is not specified proxy caching is disabled.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED on success.
	If this handler was invoked for a phase other than HANDLE then PIAPI_ERROR
	is returned.

Note:
Example:
	<PRE>
	&ltObject&gt
		Name BasicProxy
		Class BasicProxyClass
	&lt/Object&gt

	&ltObject&gt
		...
		Handle BasicProxy
		...
	&lt/Object&gt

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_IOOBJECT	"IOObject"
#define KEY_CONF_CACHEROOT	"CacheRoot"
#define KEY_CONF_ALLOWHOSTS	"AllowHosts"

#define FIELD_LENGTH 		(127)
#define CACHE_PATH_LENGTH 	(511)

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class CacheRecord
{
public:
	int iVersion;
	char szContentType[FIELD_LENGTH+1];
	char szContentEncoding[FIELD_LENGTH+1];
	char szFileName[CACHE_PATH_LENGTH+1];
	time_t tLastModified;

	CacheRecord()
	:	iVersion(1)
		{
		memset( szContentType, 0, FIELD_LENGTH );
		memset( szContentEncoding, 0, FIELD_LENGTH );
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class BasicProxy : public HandlerBaseHTTP
{
private:
	/* ---
	Configuration data
	--- */
	PIObject *pIO;		/* --- prototype IO object for proxy connections --- */
	PIString sCacheRoot;
	DblList lAllowHosts;

	/* --- forbid copy constructor --- */
	BasicProxy( const BasicProxy &t )
	: HandlerBaseHTTP( t )
		{ assert( 0 ); };

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "BasicProxy: ";
		if ( !PIUtil_stricmp( KEY_CONF_ALLOWHOSTS, pVariable ) )
			{
			lAllowHosts.Append( (DblList::type)
				PI_NEW( PIString( pValue ) ) );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_CACHEROOT, pVariable ) )
			{
			Pi3String *pTmp = Pi3String_new( pValue );
			HTTPCore_relativeToAbsolutePath( PIObject_getDB( Object() ),
                pValue, pTmp );
			sCacheRoot = Pi3String_getPtr( pTmp );
			Pi3String_delete( pTmp );
			int i=sCacheRoot.Len();
			if ( i )
				{
				const char *pTmp = sCacheRoot;
				char cLast = pTmp[i-1];
				if ( cLast!='/' && cLast!='\\' )
					{ sCacheRoot.Concatenate( HTTP_DIRSEPERATOR ); };
				};
			}
		else if ( !PIUtil_stricmp( KEY_CONF_IOOBJECT, pVariable ) )
			{
			if ( pIO )
				{
				os << "'IOObject' may only be specified once." << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};

			PIObject *pTmpIO = PIObject_loadFromLine( 
				PIObject_getDB( Object() ),
				PIObject_getConfigurationDB( Object() ),
				pValue );

			if ( !pTmpIO )
				{ return 0; };

			pIO = pTmpIO;
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	BasicProxy( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject ), pIO( 0 )
		{
		ReadParameters( iArgc, ppArgv );
		if ( !IsOK() )
			{ return; };

		if ( !pIO )
			{
			CONFIG_ERR( Object(), "BasicProxy: Missing definition for \
'IOObject'." );
			SetOK( 0 );
			return;
			};
		
		assert( pIO );
		};

	~BasicProxy()
		{
		for( DblListIterator i(lAllowHosts); !i.BadIndex(); i++ )
			{ PI_DELETE( (PIString *)i.Current() ); };
		PIObject_delete( pIO, 0, 0 );
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\

							Utilities

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	/* ---
	Filesystem canonical name given a name with (possibly) filesystem
	excluded characters
	--- */
	void Canonical( const PIString &sRawString, PIString &sResult )
		{
		Pi3String *pRawString = Pi3String_new( sRawString );
		Pi3String *pTmp = Pi3String_new( 0 );
		HTTPUtil_urlEncode( pRawString, pTmp );
		sResult = Pi3String_getPtr( pTmp );
		Pi3String_delete( pTmp );
		};

	/* ---
	Get the full path to a cached file
	--- */
	void CacheFileName( const char *pHost, const char *pPort, 
		const char *pURI, const char *pExtn, PIString &sFileName )
		{
		assert( pHost );
		assert( pURI );
		assert( pExtn );
		
		sFileName = sCacheRoot;
		PIString sTmp( pHost );
		if ( pPort )
			{
			sTmp.Concatenate( ':' );
			sTmp.Concatenate( pPort );
			};
		Pi3String *pTmp = Pi3String_new( sTmp );
		Pi3String *pTmp2 = Pi3String_new( 0 );
		HTTPUtil_urlEncode( pTmp, pTmp2 );
		sFileName.Concatenate( Pi3String_getPtr( pTmp2 ) );
		sFileName.Concatenate( '_' );
		Pi3String *pTmp3 = Pi3String_new( pURI );
		HTTPUtil_urlEncode( pTmp3, pTmp2 );
		sFileName.Concatenate( Pi3String_getPtr( pTmp2 ) );
		sFileName.Concatenate( '.' );
		sFileName.Concatenate( pExtn );
		Pi3String_delete( pTmp );
		Pi3String_delete( pTmp2 );
		Pi3String_delete( pTmp3 );
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\

								Guts

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	/* ---
	GetProxyConnection(), get a connection to the proxy server, either
	by creating a new one or by reusing an existing connection, kept
	open from a previous request
	--- */
	PIObject *GetProxyConnection( PIHTTP &tPIHTTP, const char *pProtocol,
		const char *pHost, int iPort )
		{
		assert( pProtocol );
		assert( pHost );

		(void)tPIHTTP;

		/* ---
		## NOTE ##, connection reuse is not used right now, later, a mutex
		should protect access to a list of connection resources and an 
		existing connection with the same protocol, host and port tuple
		should be returned if available

		Right now a new connection is created every time around.
		--- */

		/* --- only support HTTP proxies right now --- */
		if ( PIUtil_stricmp( pProtocol, "http" ) )
			{ return 0; };
		
		/* --- establish a connection to the remote server --- */
		const char *ppConnectParameters[2];
		ppConnectParameters[0] = pHost;
		ppConnectParameters[1] = (const char *)iPort;
		PIObject *pConn = PIObject_copy( pIO, 2, ppConnectParameters );
		return pConn;
		};
	
	/* ---
	ReleaseProxyConnection(), if we are managing a pool of connections then
	return this connection to the pool if necessary, the flag bDelete, can
	be set to indicate that headers are some other means have indicated that
	this connection should not be kept open
	--- */
	void ReleaseProxyConnection( PIObject *pConnection, int bDestroy )
		{
		(void)bDestroy;

		/* --- pool not implemented, always delete connection --- */
		if ( pConnection )
			{ PIObject_delete( pConnection, 0, 0 ); };
		};

	/* ---
	Attempts to satisfy a request by using cached data
		returns:
			0 indicates request not satisfied
			1 indicates request satisified from cache
			-1 indicates error
	--- */
	int HandleFromCachedData( PIHTTP &tPIHTTP, const PIString &sMetaFileName )
		{
		PIDB *pR = tPIHTTP.pResponseDB;
		PIDB *pQ = tPIHTTP.pRequestDB;

		PIFInfo *pFInfo = HTTPCore_getCachedFile( sMetaFileName );

		if ( !pFInfo || !PIFInfo_exists( pFInfo ) )
			{
			return 0;
			};

		int iRet = !( PIFInfo_getSize( pFInfo ) > 0 );

		CacheRecord tRecord;

		/* --- get filemap object --- */
		const char *pMap = 0;
		if ( !iRet )
			{
			int iLen;
			PIFMap *pFileMap = PIFMap_new( pFInfo );
			if ( pFileMap )
				{
				pMap = (const char *)PIFMap_lock( pFileMap, &iLen );
				if ( !pMap )
					{
					iRet = -1;
					}
				else
					{
					if ( iLen>=((int)sizeof( CacheRecord )) )
						{
						memcpy( &tRecord, pMap, sizeof( CacheRecord ) );
						}	
					else
						{
						iRet = -1;
						};
					};
				PIFMap_delete( pFileMap );
				};
			};
	
		HTTPCore_releaseCachedFile( pFInfo );	

		/* ---
		Report errors from getting the cache record
		--- */
		if ( iRet )
			{
			HTTPCore_logError( &tPIHTTP, "BscProxy: Could not get \
cache meta file with path '%s'.", (const char *)sMetaFileName );
			return -1;
			};

		/* --- now get the data file --- */
		pFInfo = HTTPCore_getCachedFile( tRecord.szFileName );
		if ( !pFInfo )
			{
			HTTPCore_logError( &tPIHTTP, "BscProxy: Could not open \
cache data file with path '%s'.", tRecord.szFileName );
			};

		/* --- add headers --- */
		Pi3String *pTmp = Pi3String_new( 0 );
		PIPlatform_beforeUnsafeBlock();
		struct tm *pTms = gmtime( &( tRecord.tLastModified ) );
		if ( pTms )
			{
			HTTPUtil_rFC822Time( pTms, pTmp );
			};

		PIPlatform_afterUnsafeBlock();
		PIString sTmp = Pi3String_getPtr( pTmp );
		Pi3String_delete( pTmp );
		if ( sTmp.Len() )
			{ PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_LASTMODIFIED,
				(void *)(const char *)sTmp, 0 );
			};
		
		if ( *( tRecord.szContentType ) )
			{ PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE,
				tRecord.szContentType, 0 );
			};

		if ( *( tRecord.szContentEncoding ) )
			{ PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTENCODING,
				tRecord.szContentEncoding, 0 );
			};

		enum { BUF_SIZE=63 };
		char szBuf[BUF_SIZE+1];
		sprintf( szBuf, "%ld", (long)PIFInfo_getSize( pFInfo ) );
		PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, szBuf, 0 );

		tPIHTTP.iStatus = ST_OK;

		/* ---
		Start response
		--- */
		/* --- start response --- */
		if (	HTTPCore_sendGeneralHeaders( &tPIHTTP ) ||
				HTTPCore_sendEntityHeaders( &tPIHTTP, pR ) )
			{
			HTTPCore_releaseCachedFile( pFInfo );	
			HTTPCore_logError( &tPIHTTP, "BscProxy: Error sending \
response headers for cached data file '%s'.", tRecord.szFileName );
			return -1;
			};

		/* ---
		Send data 
		--- */
		int iMethod = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE,
			KEY_HTTP_METHOD, 0 );
		switch( iMethod )
			{
			case MD_GET:
				HTTPUtil_sendFile( tPIHTTP.pBuffer, pFInfo, PIIOBUF_NONE, 0 );
				break;
		
			case MD_HEAD:
				break;

			default:
				assert( 0 );
				HTTPCore_releaseCachedFile( pFInfo );	
				return -1;
			};

		HTTPCore_releaseCachedFile( pFInfo );	

		/* --- sent --- */
		return 1;
		};

	/* ---
	Create cache record
		returns:
			1, file created
			0, file already exists
			-1, error 
	--- */
	int CreateCacheRecord( PIHTTP &tPIHTTP, const PIString &sMetaFileName,
			const PIString &sFileName )
		{
		PIDB *pR = tPIHTTP.pResponseDB;

		CacheRecord tRecord;

		/* --- get file meta data --- */
		const char *pCT = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
			KEY_HTTP_CONTENTTYPE, 0 );
		const char *pCE = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
			KEY_HTTP_CONTENTENCODING, 0 );
		const char *pLM = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
			KEY_HTTP_LASTMODIFIED, 0 );
		const char *pDate = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
			KEY_HTTP_DATE, 0 );

		/* --- localize the last modified time to this servers time --- */
		struct tm tLM;
		time_t tLocalLM;
		time_t tT;
		time( &tT );
		if ( pLM && HTTPUtil_readTime( pLM, &tLM ) )
			{
			time_t tFile = mktime( &tLM );
			struct tm tDate;
			if ( pDate && HTTPUtil_readTime( pDate, &tDate ) )
				{
				/* --- server sent its time, calculate localized LM time --- */
				time_t tRemoteTime = mktime( &tDate );
				tLocalLM = tFile + ( tT - tRemoteTime );
				}
			else
				{
				/* --- server didn't send its time, trust the LM date --- */
				tLocalLM = tFile;
				};

			/* --- increment LM time by 1 second to avoid sync. problems --- */
			tLocalLM++;
			}
		else
			{ tLocalLM = (time_t)-1; };

		/* --- set record data --- */
		if ( pCT )
			{ strncpy( tRecord.szContentType, pCT, FIELD_LENGTH ); };
		if ( pCE )
			{ strncpy( tRecord.szContentEncoding, pCE, FIELD_LENGTH ); };
		tRecord.tLastModified = tLocalLM;
		strncpy( tRecord.szFileName, sFileName, CACHE_PATH_LENGTH );
		tRecord.szFileName[CACHE_PATH_LENGTH] = '\0';

		/* --- attempt to create file --- */
		PIPLATFORM_FD tFd = PIFile_open( sMetaFileName, "wn" );
		if ( tFd==PIPLATFORM_FD_INVALID )
			{
#if 0
/* 
** NOTE: In this case delete all files to allow them to be created cleanly
** next time
*/
			/* --- meta file created, another process/thread got there --- */
			return 0;
#endif
			remove( sMetaFileName );
			return -1;
			};
		int iRet = PIFile_write( tFd, sizeof( tRecord ), &tRecord );
		if ( iRet )
			{
			/* --- could not write file --- */
			remove( sMetaFileName );
/////// XXXX Fd LATER
			return -1;
			};

/////// XXXX Fd LATER
		/* --- cache file created --- */
		return 1;
		};

	/* ---
	Write output from one buffer into another. Transfer data until no more
	data can be read.
	Returns 0 on success.
	--- */
	int WriteBufferToBuffer( PIIOBuffer &tInBuffer, PIIOBuffer &tOutBuffer )
		{
		int iLen;
		const char *pBuffer = PIIOBuffer_read( &tInBuffer, &iLen );
		while( pBuffer )
			{
			int iRet = PIIOBuffer_write( &tOutBuffer, pBuffer, iLen,
				PIIOBUF_NOBUFFER );
			switch( iRet )
				{
				case 0: return 0;
				case -1: return -1;
				default:
					assert( iRet>0 );
				};
			pBuffer = PIIOBuffer_read( &tInBuffer, &iLen );
			};
		return 0;
		};

	/* ---
	Write output from one buffer into another.
	iReadLen gives number of bytes to transfer, must be > 0
	Returns 0 on success.
	--- */
	int WriteBufferToBuffer( PIIOBuffer &tInBuffer, PIIOBuffer &tOutBuffer,
		int iReadLen )
		{
		assert( iReadLen>0 );
		if ( iReadLen<=0 )
			{ return 0; };
		enum { BUF_SIZE=4096 };
		char szBuffer[BUF_SIZE];
		int iAttemptToRead = (iReadLen >= BUF_SIZE) ? BUF_SIZE : iReadLen;

		int iLen;
		iLen = PIIOBuffer_readToBuffer( &tInBuffer, szBuffer, iAttemptToRead );

		while( iLen>0 )
			{
			if ( PIIOBuffer_write( &tOutBuffer, szBuffer, iLen, PIIOBUF_NOBUFFER )
				<=0 )
				{ return -1; };
			iAttemptToRead -= iLen;
			if ( !iAttemptToRead )
				{ return 0; };

			iLen = PIIOBuffer_readToBuffer( &tInBuffer, szBuffer,
				iAttemptToRead );
			};

		/* ---
		Done
		--- */
		return iLen;
		};

	/* ---
	Top level proxy server stuff
	--- */
	int DoProxy( PIHTTP &tPIHTTP )
		{
		assert( pIO );

		PIDB *pR = tPIHTTP.pResponseDB;
		PIDB *pQ = tPIHTTP.pRequestDB;

		/* --- get the hostname, port and protocol --- */
		const char *pHost = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			KEY_HTTP_PROXYHOST, 0 );
		const char *pPort = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			KEY_HTTP_PROXYPORT, 0 );
		int iPort = ( pPort && *pPort ) ? atoi( pPort ) : HTTP_DEFAULT_PORT;
		const char *pProto = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			KEY_HTTP_PROXYPROTOCOL, 0 );
		if ( !pHost || !pProto || iPort<0 )
			{
			/* --- this is an error --- */
			HTTPCore_logError( &tPIHTTP, "BasicProxy: Invalid host '%s', \
port '%d' or protocol '%s'.", pHost ? pHost : "", iPort, pProto ? pProto : "" );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_BADREQUEST );
			};

		/* ---
		Check if the current host corresponds to a host in the allow list
		--- */
		int iFound = 0;
		if ( !lAllowHosts.Size() )
			{ iFound = 1; }
		else
			{
			for( DblListIterator iAllowHosts( lAllowHosts );
				!iAllowHosts.BadIndex(); iAllowHosts++ )
				{
				PIString &sTmp = *( (PIString *)iAllowHosts.Current() );
				if ( HTTPUtil_regexMatch( sTmp, sTmp.Len(), pHost, 
					strlen( pHost ) ) )
					{
					iFound = 1;
					break;
					};
				};
			};
		if ( !iFound )
			{
			/* ---
			list of allowed hosts was specified but this host did
			not match any of them
			--- */
			tPIHTTP.iStatus = ST_FORBIDDEN;	
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_FORBIDDEN );
			};

		/* --- get the method and URL path and Query String --- */
		const char *pMethod = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			KEY_HTTP_METHOD, 0 );
		int iMethod = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_METHOD,
			0 );
		const char *pURI = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			KEY_HTTP_URI, 0 );
		const char *pQS = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			KEY_HTTP_QUERYSTRING, 0 );
		if ( !pMethod || !pURI )
			{
			HTTPCore_logError( &tPIHTTP, "BasicProxy: Internal error" );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* ---
		Generate filename of meta file
		--- */
		PIString sMetaFileName;
		CacheFileName( pHost, pPort, pURI, "met", sMetaFileName );

		/* ---
		See if the request can be correctly handled by using cached data
		--- */
        if ( !pQS )
            {
			switch( HandleFromCachedData( tPIHTTP, sMetaFileName ) )
				{
				case 1:		return PIAPI_COMPLETED;
				case 0:		break;
				default:
					tPIHTTP.iStatus = ST_INTERNALERROR;
					return PIAPI_COMPLETED;
				};
            };

		/* ---
		Get a connection to the server, either by creating a new one or
		by reusing an existing connection
		--- */
		PIObject *pConn = GetProxyConnection( tPIHTTP, pProto, pHost,
			iPort );

		if ( !pConn )
			{
			/* --- this is an error --- */
			HTTPCore_logError( &tPIHTTP, "BasicProxy: Failed to establish IO \
connection to remote server with location '%s://%s:%d'.", pProto, pHost,
				iPort );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* --- make a buffer object --- */
		PIIOBuffer *pProxy = PIIOBuffer_new( pConn );

		/* ---
		Issue the HTTP request
		--- */

		/* --- place status line in buffer --- */
		PIIOBuffer_write( pProxy, pMethod, -1, PIIOBUF_NONE );
		PIIOBuffer_write( pProxy, " ", 1, PIIOBUF_NONE );
		PIIOBuffer_write( pProxy, pURI, -1, PIIOBUF_NONE );
        if ( pQS && *pQS )
            {
			PIIOBuffer_write( pProxy, "?", 1, PIIOBUF_NONE );
			PIIOBuffer_write( pProxy, pQS, -1, PIIOBUF_NONE );
            };
		PIIOBuffer_write( pProxy, " ", 1, PIIOBUF_NONE );
		PIIOBuffer_writeLn( pProxy, "HTTP/1.0", 8, PIIOBUF_NONE );

		/* ---
		Modify client headers 
		--- */
/*
** NOTE: Reexamine this later
**
**		PIDBIterator *pIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822,
**			KEY_HTTP_CONNECTION, 0 );
**		if ( pIter )
**			{
**			while( PIDBIterator_atValidElement( pIter ) )
**				{ PIDBIterator_removeCurrent( pIter ); };
**			PIDBIterator_delete( pIter );
**			};
*/

		/* --- iterate over all headers, sending them to server --- */
		PIDBIterator *pRIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822, 0, 0 );
		if ( pRIter )
			{
			for( ;	PIDBIterator_atValidElement( pRIter );
					PIDBIterator_next( pRIter ) )
				{
				const char *pKey;
				const char *pValue = (const char *)
					PIDBIterator_current( pRIter, &pKey );
				if ( pKey && *pKey )
					{
					PIIOBuffer_write( pProxy, pKey, -1, PIIOBUF_NONE );
					PIIOBuffer_write( pProxy, ": ", 2, PIIOBUF_NONE );
					PIIOBuffer_writeLn( pProxy, pValue, -1, PIIOBUF_NONE );
					};
				};
			PIDBIterator_delete( pRIter );
			};
		PIIOBuffer_writeLn( pProxy, "", 0, PIIOBUF_NONE );

		/* ---
		Send input from client, if any
		--- */
		const char *pContentLength = (const char *)PIDB_lookup( pQ,
			PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0 );
		int iStdinLen = pContentLength ? atoi( pContentLength ) : 0;
		if ( iStdinLen>0 )
			{
			if ( WriteBufferToBuffer( *( tPIHTTP.pBuffer ), *pProxy,
				iStdinLen )<0 )
				{	
				/* --- this is an error --- */
				HTTPCore_logError( &tPIHTTP, "BasicProxy: Premature close \
on proxy connection to location '%s://%s:%d'.", pProto, pHost, iPort );
				PIIOBuffer_delete( pProxy );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			};

		/* --- send the whole thing on its way --- */
		PIIOBuffer_flush( pProxy );

		/* ---
		Read the response status line and headers
		--- */
		int iRet = HTTPCore_readHeaders( pProxy, pR, RH_RESPONSE );
		if ( iRet )
			{
			HTTPCore_logError( &tPIHTTP, "BasicProxy: Error reading \
response headers from remote server. Error code is %d.", iRet );
			iRet = -1;
			};

		/* ---
		Check out status line from remote server
		--- */
		const char *pCLF = 0;
		if ( !iRet && ( pCLF = (const char *)
			PIDB_lookup( pR, PIDBTYPE_STRING, KEY_HTTP_CLF, 0 ) ) ) 
			{
			/* ---
			Get response protocol
			--- */
			/* ?? */

			/* ---
			Get response content type
			--- */
			const char *pRContentType = (const char *)PIDB_lookup( pR,
				PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, 0 );

			/* ---
			Get response content length
			--- */
			const char *pRContentLength = (const char *)PIDB_lookup( pR,
				PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0 );
			int iRContentLength = pRContentLength ? atoi( pRContentLength ) : 0;

			/* ---
			Get response pragma
			--- */
			const char *pRPragma = (const char *)PIDB_lookup( pR,
				PIDBTYPE_RFC822, KEY_HTTP_PRAGMA, 0 );

			/* ---
			Get expires header
			--- */
			const char *pExpires = (const char *)PIDB_lookup( pR,
				PIDBTYPE_RFC822, KEY_HTTP_EXPIRES, 0 );

			/* ++++ __________________________ +++ *
						Got status line
			 * ++++ __________________________ +++ */
			/* --- grap the status code sent by the CGI --- */
			int i=0;
			for(; pCLF[i] && !(isspace(pCLF[i])); i++); /* advance to space */
			for(; pCLF[i] && (isspace(pCLF[i])); i++); /* advance past space */
			tPIHTTP.iStatus = atoi( &(pCLF[i]) );

			/* ---
			See if we want to cache this response. Most of the cachable
			responses will be simple requests for content. Although more
			complex requests could be handled correctly, its a better 
			'quick win' to only handle the simplest of requests which meet the
			following criteria and forward the entire response from other
			requests back to the client.
			--- */
			PIString sMessageMedia( "*essage/*" );
			PIString sRContentType( pRContentType );
			if ( 							/* method must be 'GET' or 'HEAD' */
				( iMethod==MD_GET || iMethod==MD_HEAD ) &&
				( !HTTPUtil_regexMatch(
						sMessageMedia, sMessageMedia.Len(),
						sRContentType, sRContentType.Len()
						) ) &&
											/* no 'message' types, 
											including multipart */
				iRContentLength>0 &&		/* sane content length  */
                !pQS &&                     /* no query string */
											/* no 'Pragma: no-cache' */
				( !pRPragma || PIUtil_strncmpi( pRPragma, "no-cache", 8 ) ) &&
				!pExpires &&				/* no expires header */
				sCacheRoot.Len() )
				{
				PIString sFileName;
				CacheFileName( pHost, pPort, pURI, "dat", sFileName );
				PIPLATFORM_FD tFD;
				tFD = PIFile_open( sFileName, "w" );
				if ( tFD != PIPLATFORM_FD_INVALID )
					{
					int iLen;
					const char *pBuffer = PIIOBuffer_read( pProxy, &iLen );
					while( pBuffer && !iRet )
						{
						if ( PIFile_write( tFD, iLen, pBuffer ) )
							{
							/* --- error --- */
							HTTPCore_logError( &tPIHTTP, "BasicProxy: \
Error writing file to cache." );
							iRet = -1;
							};
						pBuffer = PIIOBuffer_read( pProxy, &iLen );
						};
					PIFile_close( tFD );

					/* ---
					Attempt to create cache record
					--- */
					switch( CreateCacheRecord( tPIHTTP, sMetaFileName,
						sFileName ) )
						{
						case 0:
							/* --- fall through --- */
						case 1:
							/* --- remove response RFC822 headers --- */
							{
							PIDBIterator *pIter = PIDB_getIterator( pR,
								PIDBTYPE_RFC822, 0, 0 );
							if ( pIter )
								{
								while( PIDBIterator_atValidElement( pIter ) )
									{ PIDBIterator_removeCurrent( pIter ); };
								PIDBIterator_delete( pIter );
								};
							};

							/* --- issue redirect to send cached data --- */
							PIIOBuffer_delete( pProxy );
							ReleaseProxyConnection( pConn, 0 );
							return INT_REDIRECT;

						default:
							/* --- error --- */
							remove( sFileName );
							HTTPCore_logError( &tPIHTTP, "BasicProxy: \
Error writing meta file to cache." );
							iRet = -1;
						};
					};
				};

			if ( iRet )
				{
				PIIOBuffer_delete( pProxy );
				ReleaseProxyConnection( pConn, 0 );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};	

			/* ---
			Right now disable keep alive, because we don't trust the 
			headers returned from the server (in terms of content length)
			--- */
			PIDB_replace( tPIHTTP.pConnectionDB, PIDBTYPE_OPAQUE,
				KEY_INT_KEEPOPEN, (void *)0, 0 );

			/* --- start response --- */
			if (	HTTPCore_sendGeneralHeaders( &tPIHTTP ) ||
					HTTPCore_sendEntityHeaders( &tPIHTTP, pR ) )
				{
				iRet = -1;
				};

			if ( !iRet )
				{
				/* ---
				Send response from server back to client
				--- */
				if ( WriteBufferToBuffer( *pProxy, *( tPIHTTP.pBuffer ) )<0 )
					{
					HTTPCore_logError( &tPIHTTP, "BasicProxy: Premature close \
from remote HTTP server '%s://%s:%d'.", pProto, pHost, iPort );
					iRet = -1;
					};
				};
			}
		else
			{
			HTTPCore_logError( &tPIHTTP, "BasicProxy: Bad response \
from remote HTTP server '%s://%s:%d'.", pProto, pHost, iPort );
			iRet = -1;
			};

		/* --- send the entire result back to the client --- */

		/* --- done --- */
		PIIOBuffer_delete( pProxy );
		ReleaseProxyConnection( pConn, 0 );
		if ( iRet )
			{
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};	
		return PIAPI_COMPLETED;
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\

						Standard Handle Function

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &/* tBuffer */ )
		{
		if ( iPhase!=PH_HANDLE )
			{ return PIAPI_ERROR; };

		return DoProxy( tPIHTTP );
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int BasicProxy_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( BasicProxy( pObj,
		iArgc, ppArgv ) ) );
}

