/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: Headers.cpp$
 * $Date: Sun Aug 10 06:36:48 1997$
 *
 Description:
\*____________________________________________________________________________*/
//$SourceTop:$

#include "HandBase.h"
#include "HTTPCore.h"
#include "PIStrStr.h"
#include "HTTPDefs.h"

/*____________________________________________________________________________*\
 *
 Description:
	Documentation
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	ReadRequestHeaders

Description:
	Read the status line and client request headers for a new HTTP client
	request.
	On read a correct request this handler may disable connection keep open
	for the request if it is enable and any of the following conditions are
	met:-
<UL>
<LI>The client protocol is HTTP/1.0 and the 'Connection: Keep-Alive' header 
variable-value pair does not exist.
<LI>The client protocol is HTTP/1.1 and the 'Connection: Close' header
variable-value pair exists.
</UL>

<!--
 no options

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>RequirePermissions
<TD>-
<TD>Combination of 'X', 'R', 'X'
<TD>Permissions required on file
<TD>RequirePermissions="X"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	Require Permissions
</H5>
These permissions will be compared against the permissions on the filesystem
resource to determine if this handler will indicate a match by returning
PIAPI_COMPLETED.
This value is obtained by concatenating any of the following flags together.
<CENTER>
<TABLE BORDER=1>
<TH>Option
<TH>Meaning
<TR>
<TD>R
<TD>Resource must be readable
<TR>
<TD>W
<TD>Resource must be writable
<TR>
<TD>X
<TD>Resource must be executable
</TABLE>
</CENTER>
Permission flags are not case sensitive.

-->

Phase:
	HEADERS

Returns:
	PIAPI_COMPLETED if the phase completed without serious error. An HTTP
	error status response (such as '400 Bad Request') may have been set.
	On error PIAPI_ERROR is returned. An example of an error would be 
	invoking this handler for a phase other that 'HEADERS'.

Example:
	<PRE>
	&ltObject&gt
		Name ReadRequestHeaders
		Class ReadRequestHeadersClass
	&lt/Object&gt

	&ltObject&gt
		...
		Headers ReadRequestHeaders
		...
	&lt/Object&gt

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ReadRequestHeaders : public HandlerBaseHTTP
{
private:
	/* ---
	Configuration data
	--- */

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "ReadRequestHeaders: ";
/* No parameters */
/*		if ( !PIUtil_stricmp( KEY_CONF_REQUIREPERMISSIONS, pVariable ) ) */
		if ( 0 )
			{
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	ReadRequestHeaders( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject )
		{
		ReadParameters( iArgc, ppArgv );
		};

	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &tB )
		{
		if ( iPhase!=PH_HEADERS ) { return PIAPI_ERROR; };

		PIDB *pC = tPIHTTP.pConnectionDB;
		PIDB *pQ = tPIHTTP.pRequestDB;

		int iStatus = HTTPCore_readHeaders( &tB, pQ, RH_CLIENTREQUEST );

		/* --- get keep open flag --- */
		int iKeepOpen = (int)PIDB_lookup( pC, PIDBTYPE_OPAQUE,
			KEY_INT_KEEPOPEN, 0 );

		if ( iStatus==INT_CONTINUE )
			{
			/* ---
			If the server has allowed connections to be kept open the
			keep open flag will be set. If this is set then determine from
			the client request headers if the connection should be
			kept open
			--- */
			if ( iKeepOpen )
				{
				/* --- get protocol --- */
				int iProtocolNumber = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE,
					KEY_HTTP_PROTOCOL, 0 );

				/* --- get the Connection header from the client --- */
				const char *pConnection = (const char *)PIDB_lookup(
					pQ, PIDBTYPE_RFC822, KEY_HTTP_CONNECTION, 0 );	
				switch( iProtocolNumber )
					{
					case PR_HTTP10:
						/* --- HTTP/1.0 --- */
						if ( !pConnection ||
							PIUtil_strncmpi( pConnection, "Keep-Alive", 10 ) )
							{
							PIDB_replace( pC, PIDBTYPE_OPAQUE,
								KEY_INT_KEEPOPEN, 0, 0 );
							};
						break;

					case PR_HTTP11:
						/* --- HTTP/1.1 --- */
						if ( pConnection &&
							!PIUtil_strncmpi( pConnection, "Close", 5 ) )
							{
							PIDB_replace( pC, PIDBTYPE_OPAQUE,
								KEY_INT_KEEPOPEN, 0, 0 );
							};
						break;

					default:;
						PIDB_replace( pC, PIDBTYPE_OPAQUE,
							KEY_INT_KEEPOPEN, 0, 0 );
					};
				};
			}
		else
			{
			PIDB_replace( pC, PIDBTYPE_OPAQUE, KEY_INT_KEEPOPEN, 0, 0 );
			};

		switch( iStatus )
			{
			case PIAPI_COMPLETED:
				return PIAPI_COMPLETED;

			default:
				tPIHTTP.iStatus = iStatus;
				return PIAPI_COMPLETED;
			};
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int ReadRequestHeaders_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( ReadRequestHeaders( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name ReadRequestHeadersClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor ReadRequestHeaders_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name ReadRequestHeaders
		Class ReadRequestHeadersClass
	</Object>

/*___+++CNF_END+++___*/
#endif

