/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: Admin.cpp$
 * $Date: Sun Aug 10 06:40:13 1997$
 *
 Description:

\*____________________________________________________________________________*/
/* $SourceTop:$ */

#include <assert.h>
#include <fstream.h>
#include <ctype.h>
#include <stdio.h>

#include "QConfig.h"
#include "Pi3File.h"
#include "HTTPCore.h"
#include "PIStrStr.h"
#include "Admin.h"
#include "PIStream.h"

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
class _Admin
{
private:
    int iRemote;

public:
    ConnectionPool *pPool;
    PIString sFileName;
    PIString sHost;
    int iPort;
    PIString sUsername;
    PIString sPassword;

private:
    int iOK;

public:
    _Admin(
        PIObject *pParent,
        const char *pHost,
        int iThePort,
        const char *pUsername,
        const char *pPassword )
    :
        iRemote( 1 ),
        pPool( 0 ),
        sFileName(),
        sHost( pHost ),
        iPort( iThePort ),
        sUsername( pUsername ),
        sPassword( pPassword ),
        iOK( 0 )
        {
        HTTPCore_initClient();

        PIObject *pIOObject = PIObject_loadFromLine(
            PIObject_getDB( pParent ),
            PIObject_getConfigurationDB( pParent ),
            "ClientIOObject" );

        if ( !pIOObject )
            { return; };

        pPool = ConnectionPool_new( pIOObject, 0 );

        iOK = 1;
        };

    _Admin( const char *pFileName )
    :
        iRemote( 0 ),
        pPool( 0 ),
        sFileName( pFileName ),
        sHost(),
        iPort( -1 ),
        sUsername(),
        sPassword(),
        iOK( 1 )
        {};

    inline int IsOK() const         { return iOK; };
    inline int IsRemote() const     { return iRemote; };

    ~_Admin()
        {
        if ( pPool )
            { ConnectionPool_delete( pPool ); };

        if ( IsRemote() )
            { HTTPCore_cleanupClient(); };
        };
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int Internal_loadConfigFromFile( QConfig *pConfig,
    const char *pFileName )
{
	assert( pFileName );

	fstream ifs( pFileName, ios::in | ios::nocreate );
	if ( !ifs || ifs.eof() )
		{ return -1; };
	
    int iRet = Pi3File_readConfig( pConfig, ifs );

    ifs.close();

	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int Internal_writeConfigToFile( QConfig *pConfig,
    const char *pFileName )
{
	assert( pFileName && pConfig );
	ofstream ofs( pFileName, ios::out );

	if ( ofs.bad() || ofs.eof() )
		{ return -1; };

    int iRet = Pi3File_writeConfig( pConfig, ofs );

	ofs.close();

	return iRet;	/* success */
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
    Returns a PIAPI succes code or an HTTP request status, in any case
    0 on success, otherwise an error.
\*____________________________________________________________________________*/
static int Internal_makeHTTPRequest(
    PIIOBuffer *pBuffer,
    int iMethod,
    const char *pAdminCommand,
    const char *pHost,
    int iPort,
    const char *pContentType,
    const char *pContentLength,
    int (* fnPutData)( void *pParam ), void *pPutDataParam,
    int (* fnGetData)( void *pParam ), void *pGetDataParam,
    PIString &sError
    )
{
    assert( pBuffer );

    const char *pCLF;

    PIDB *pQ = PIDB_new( 0, "Q" );
    PIDB *pR = PIDB_new( 0, "R" );

    /*
    ** Add some headers
    */ 
    PIDB_add( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_METHOD, (void *)iMethod, 0 );
    PIDB_add( pQ, PIDBTYPE_STRING, KEY_HTTP_URI, "/", 0 );
    PIDB_add( pQ, PIDBTYPE_RFC822, KEY_HTTP_X_ADMINCOMMAND,
        (void *)pAdminCommand, 0 );
    PIDB_add( pQ, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE,
        (void *)pContentType, 0 );
    PIDB_add( pQ, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH,
        (void *)pContentLength, 0 );

    /* ---
    Make sure the 'host' header is sent
    --- */
    PIOStrStream os;
    os << pHost;
    if ( iPort != HTTP_DEFAULT_PORT )
        {
        os << ":" << iPort;
        };
    os << ends;
    PIDB_replace( pQ, PIDBTYPE_RFC822, KEY_HTTP_HOST,
        (void *)os.str(), 0 );

    /*
    ** Make the request
    */
    int iRet = HTTPUtil_doHTTPRequest( pBuffer, pQ );
    if ( iRet != PIAPI_COMPLETED )
        {
        sError = "Failed to connect to remote server '";
        sError.Concatenate( pHost );
        sError.Concatenate( "'." );
        return iRet;
        };

    /* --- more data to send? --- */
    if ( fnPutData!=0 )
        {
        iRet = (fnPutData)( pPutDataParam );
        if ( iRet )
            {
            sError = "Error sending data to remote server '";
            sError.Concatenate( pHost );
            sError.Concatenate( "'." );
            return iRet;
            };
        };

    /* --- send the whole thing on its way --- */
    PIIOBuffer_flush( pBuffer );

    /*
    ** Read the response
    */
    iRet = HTTPCore_readHeaders( pBuffer, pR, RH_RESPONSE );
    if ( iRet )
        {
        sError = "Error reading response headers from remote server '";
        sError.Concatenate( pHost );
        sError.Concatenate( "'." );
        return iRet;
        };
    
    /* ---
    Check out status line from remote server
    --- */
    if ( ( pCLF = (const char *)
        PIDB_lookup( pR, PIDBTYPE_STRING, KEY_HTTP_CLF, 0 ) ) )
        {
        /* ++++ __________________________ +++ *
            Got status line
         * ++++ __________________________ +++ */
        /* --- grap the status code sent by the server --- */
        int i=0;
        for(; pCLF[i] && !(isspace(pCLF[i])); i++); /* advance to space */
        for(; pCLF[i] && (isspace(pCLF[i])); i++); /* advance past space */
        iRet = atoi( &(pCLF[i]) );

        /*
        ** Check out the status
        */
        if ( iRet>=200 && iRet<299 )
            {
            /* --- 2xx assume success under HTTP/1.0 --- */ 
            iRet = 0;
            };

        if ( iRet )
            {
            /*
            ** Set the error message
            */
            sError = "Received error response from server:\n\n";
            sError.Concatenate( &(pCLF[i]) );
            sError.Concatenate( "\n\n" );

            /*
            ** Read the rest of the response
            */
            int iLen;
            const char *pErrorBuffer = PIIOBuffer_read( pBuffer, &iLen );
            while( pErrorBuffer )
                {
                PIString sTmp( pErrorBuffer, iLen );
                sError.Concatenate( sTmp );
                pErrorBuffer = PIIOBuffer_read( pBuffer, &iLen );
                };
            };
        };

    if ( !pCLF )
        {
        sError = "Error reading response from remote server '";
        sError.Concatenate( pHost );
        sError.Concatenate( "'." );
        iRet = 400;
        };

    if ( !iRet )
        {
        /* --- expect to get data? --- */
        if ( fnGetData!=0 )
            {
            iRet = (fnGetData)( pGetDataParam );
            };
        };

	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
    All about getting a configuration file from the server.
\*____________________________________________________________________________*/
/*
** Structure for passing information to get and put callback functions
*/
struct DataParam
{
    PIIOBuffer *pBuffer;
    QConfig *pConfig;
    int iLength;
    void *pData;
    PIString tMessage;
};
static int Internal_fnGetDataFunction( void *pParam )
{
    DataParam &tParam = *( (DataParam *)pParam );
    
    /* --- make PIIStream and load the configuration file --- */
    PIIStream is( tParam.pBuffer );
    return Pi3File_readConfig( tParam.pConfig, is );
}
static int Internal_loadConfigFromServer(
    ConnectionPool *pPool,
    QConfig *pConfig,
    const char *pHost,
    int iPort,
    const char *pUsername,
    const char *pPassword,
    PIString &sError )
{
    const char *pCLF = 0;

    /*
    ** Get a connnection to the remote host from the connection pool
    */
    PIIOBuffer *pBuffer = ConnectionPool_grapConnection( pPool,
        pHost, iPort );

    if ( !pBuffer )
        {   
        return -1;
        };

    DataParam tParam;
    tParam.pBuffer = pBuffer;
    tParam.pConfig = pConfig;
    int iRet = Internal_makeHTTPRequest(
        pBuffer,
        MD_GET,
        KEY_ADMIN_GETCONFIG,
        pHost,
        iPort,
        ADMIN_CONTENTTYPE,
        "0",
        0, 0,           /* no data to put */
        Internal_fnGetDataFunction, &tParam,
        sError
        );

    /*
    ** Done with connection
    */
done:
    ConnectionPool_releaseConnection( pPool, pBuffer, true );

	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
    All about putting a configuration file onto the server.
\*____________________________________________________________________________*/
static int Internal_fnPutDataFunction( void *pParam )
{
    DataParam &tParam = *( (DataParam *)pParam );
    
    /* --- make an PIOStream and write the configuration file --- */
    return !PIIOBuffer_write( tParam.pBuffer, (const char *)tParam.pData,
        tParam.iLength, PIIOBUF_NOBUFFER );
}
static int Internal_writeConfigToServer(
    ConnectionPool *pPool,
    QConfig *pConfig,
    const char *pHost,
    int iPort,
    const char *pUsername,
    const char *pPassword,
    PIString &sError )
{
    const char *pCLF = 0;
    int iRet = 0;
    enum { BUF_SIZE=63 };
    char szFileSize[BUF_SIZE+1];
    PIFInfo *pFInfo = 0;
    int iLen;
    PIFMap *pFileMap = 0;
    const char *pMap = 0;
    DataParam tParam;
    ofstream ofs;
    int iTempFileWritten = 0;

    /*
    ** Get a connnection to the remote host from the connection pool
    */
    PIIOBuffer *pBuffer = ConnectionPool_grapConnection( pPool,
        pHost, iPort );

    if ( !pBuffer )
        {   
        return -1;
        };

    /*
    ** Open a temporary file and write the configuration into it
    */
    char szTmpFile[L_tmpnam];
    if ( !tmpnam( szTmpFile ) )
        {
        sError = "Could not create local temporary file for configuration \
upload.";
        iRet = -1;
        goto done;
        };
    ofs.open( szTmpFile );
    if ( !ofs.good() )
        {
        sError = "Could not open local temporary file '";
        sError.Concatenate( szTmpFile );
        sError.Concatenate( "' for configuration upload." );
        iRet = -1;
        goto done;
        };

    iTempFileWritten = 1;
    iRet = Pi3File_writeConfig( pConfig, ofs );
    ofs.close();
    if ( iRet )
        {
        sError = "Could write to local temporary file '";
        sError.Concatenate( szTmpFile );
        sError.Concatenate( "' for configuration upload." );
        goto done;
        };

    pFInfo = PIFInfo_new( szTmpFile );
    if ( !pFInfo )
        {
        sError = "Internal_writeConfigToServer: internal error.";
        iRet = -1;
        goto done;
        };

    sprintf( szFileSize, "%ul", PIFInfo_getSize( pFInfo ) );

    if ( PIFInfo_getSize( pFInfo )>0 )
        { pFileMap = PIFMap_new( pFInfo ); };
    if ( pFileMap )
        {
        pMap = (const char *)PIFMap_lock( pFileMap, &iLen );
        if ( pMap && iLen!=PIFInfo_getSize( pFInfo ) )
            {
            pMap = 0;
            };
        };

    /*
    ** If we got to this point without pMap, an error occurred someplace
    */
    if ( !pMap )
        {
        sError = "Could memory map local temporary file '";
        sError.Concatenate( szTmpFile );
        sError.Concatenate( "' for configuration upload." );
        iRet = -1;
        goto done;
        };

    tParam.pBuffer = pBuffer;
    tParam.pConfig = pConfig;
    tParam.iLength = iLen;
    tParam.pData = (void *)pMap;
    iRet = Internal_makeHTTPRequest(
        pBuffer,
        MD_PUT,
        KEY_ADMIN_SAVECONFIG,
        pHost,
        iPort,
        ADMIN_CONTENTTYPE,
        szFileSize,
        Internal_fnPutDataFunction, &tParam,
        0, 0,        /* no data to get */
        sError
        );

    /*
    ** Done with connection
    */
done:
    if ( pFileMap ) { PIFMap_delete( pFileMap ); };
    if ( pFInfo ) { PIFInfo_delete( pFInfo ); };
    if ( iTempFileWritten )
        {
        ::remove( szTmpFile );
        };

    ConnectionPool_releaseConnection( pPool, pBuffer, true );

	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
Admin *Admin_newLocal( const char *pFileName )
{
    return PI_NEW( _Admin( pFileName ) );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
Admin *Admin_newRemote(
    PIObject *pParent,
    const char *pHost,
    int iPort,
    const char *pUsername,
    const char *pPassword )
{
    return PI_NEW( _Admin( pParent, pHost, iPort, pUsername, pPassword ) );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void Admin_delete( Admin *pAdmin )
{
    PI_DELETE( pAdmin );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Admin_loadConfig( Admin *pAdmin, QConfig *pConfig, PIString &sError )
{
    int iRet;

    if ( !pAdmin )
        { return -1; };

    sError = PIString::Empty();

    if ( pAdmin->IsRemote() )
        {
        return Internal_loadConfigFromServer(
            pAdmin->pPool,
            pConfig,
            pAdmin->sHost,
            pAdmin->iPort,
            pAdmin->sUsername,
            pAdmin->sPassword,
            sError );
        }
    else
        {
        return Internal_loadConfigFromFile( pConfig, pAdmin->sFileName );
        };
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Admin_saveConfig( Admin *pAdmin, QConfig *pConfig, PIString &sError )
{
    int iRet;
    if ( !pAdmin )
        { return -1; };

    if ( pAdmin->IsRemote() )
        {
        return Internal_writeConfigToServer(
            pAdmin->pPool,
            pConfig,
            pAdmin->sHost,
            pAdmin->iPort,
            pAdmin->sUsername,
            pAdmin->sPassword,
            sError );
        }
    else
        {
        return Internal_writeConfigToFile( pConfig, pAdmin->sFileName );
        };
}

