/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	(http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: WinLoop.cpp$
 * $Date: Sun Aug 10 06:36:50 1997$
 *
 Description:

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>

#include "PICompat.h"
#include "Pi2API.h"
#include "IntrFace.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_MAINWINDOW         "MainWindow"

#define CONFIG_ERR(this, msg)   \
    { PILog_addMessage( PIObject_getDB(this), \
    PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
    WinLoop

Description:
    Create the top level Interface structure for windows messages 
    handling and dispatch messages on the current thread until 
    WM_QUIT is received (default windows message loop).

Example:
	<PRE>
	&ltObject&gt
		Name WinLoop
		Class WinLoopClass
	&lt/Object&gt

	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class WinLoop 
{
private:
    /* ---
    Internal data members
    --- */
    Interface *pInterface;      /* the main window dispatch structure */
    PIObject *pMainWindow;      /* the main window object */

protected:

public:
	WinLoop( PIObject *pObject, int iArgc, const char *ppArgv[] )
    :
        pInterface( 0 ),
        pMainWindow( 0 )
		{
        const char *pMainWindowName = PIConfig_lookupValue(
            PIObject_getConfigurationDB( pObject ),
            KEY_CONF_MAINWINDOW,
            0,
            0 );

        if ( !pMainWindowName )
            {
            CONFIG_ERR( pObject, "WinLoop: 'MainWindow' not defined" );
            return;
            };

        PIDB *pDB = PIObject_getDB( pObject );
        pMainWindow = PIObject_loadFromLine(
            pDB, PIObject_getConfigurationDB( pObject ), pMainWindowName );

        if ( !pMainWindow )
            {/* --- error set by PIObject_loadFromLine() --- */ return; };

        pInterface = PI_NEW( Interface ); 
		};

	~WinLoop()
		{
        PI_DELETE( pInterface ); 
        PIObject_delete( pMainWindow, 0, 0 );
		};

    int Execute( int iArgc, const char *ppArgv[] )
        {
        /*
        ** Create the main window
        */
        pInterface->iMessage = INTRFACE_MSG_WIN32CREATE;
        int iResult = PILogic_execute( pMainWindow, 1,
            (const char **)&pInterface );

        /*
        ** If the window failed to load popup the message and abort
        */
        if ( iResult != PIAPI_COMPLETED )
            {
            ::MessageBox( NULL, pInterface->sError, "Interface Creation Error",
                MB_OK );
            return PIAPI_ABORT;
            };

#if 0
        /*
        ** Do the windows loop, until WM_QUIT
        */
        MSG tMsg;
        while( ::GetMessage( &tMsg, NULL, 0, 0 ) )
            {
//            if ( (pInterface->eType==Interface::PROPSHEET) &&
//                !(PropSheet_IsDialogMessage( pInterface->hWnd, &tMsg )) )
                {
                ::TranslateMessage( &tMsg );
                ::DispatchMessage( &tMsg );
                };
            };
#endif

        return PIAPI_COMPLETED;
        };

    inline int IsOK()   { return pMainWindow!=0; };
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int WinLoop_execute( PIObject *pObj,
    int iArgc, const char *ppArgv[] )
{
    if ( !pObj ) return PIAPI_ERROR;
    return ((WinLoop *)PIObject_getUserData(pObj))->Execute( iArgc,
        ppArgv );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int WinLoop_onClassLoad( PIClass_LoadAction eLoad, int,
    const char *[] )
{
    switch( eLoad )
        {
        case STARTUP:
        default:;
        };
    return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int WinLoop_constructor( PIObject *pObj,
    int iArgc, const char *ppArgv[] )
{
    WinLoop *pWin = new WinLoop( pObj, iArgc, ppArgv );
    if (!( pWin && pWin->IsOK() ))
        {
        PI_DELETE( pWin );
        return PIAPI_ERROR;
        };

    if ( PIObject_setUserData( pObj, pWin ) )
        {
        PI_DELETE( pWin );
        CONFIG_ERR( pObj, "WinLoop_constructor: \
PIObject_setUserData() failed" );
        return PIAPI_ERROR;
        };
    return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int WinLoop_copyConstructor( PIObject *, int, const char *[] )
{
    return PIAPI_ERROR;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int WinLoop_destructor( PIObject *pObj, int,
    const char *[] )
{
    PI_DELETE( (WinLoop *)PIObject_getUserData( pObj ) );
    return PIAPI_COMPLETED;
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name WinLoopClass
		Type LogicExtension
		Library IntrFace
		OnClassLoad WinLoop_onClassLoad
		Constructor WinLoop_constructor
		CopyConstructor WinLoop_copyConstructor
		Destructor WinLoop_destructor
		Execute WinLoop_execute
	</Class>

	<Object>
		Name WinLoop
		Class WinLoopClass
        MainWindow "Window1"
	</Object>

/*___+++CNF_END+++___*/
#endif

