/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: Dispatch.cpp$
 * $Date: Sun Aug 10 06:36:29 1997$
 *
 Description:
	HTTP Dispatch class. Dispatch requests through other objects.

	TODO:
		- this file could do with a rewrite

\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include "DblList.h"
#include "PIString.h"
#include "StrToken.h"
#include "PICompat.h"

#include "Pi2API.h"
#include "PIHTTP.h"
#include "HTTPCore.h"
#include "DblList.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define DISPATCHERDB_NAME					"Dispatcher DB"
#define HANDLERSDB_NAME						"Handler DB"
#define KEY_CONF_KEEPOPEN					"KeepOpen"
#define KEY_CONF_DEFAULTHOST				"DefaultHost"
#define VALUE_OFF							"Off"

#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }
/*
#define D { cerr << __FILE__ << ", " << __LINE__ << ": " << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class Dispatcher
{
private:
	PIObject *pObject;
	PIObject *pDefaultHost;
	bool bOK;
	int iKeepOpen;
	
	/* ---
	Fast keys
	--- */
	const char *pFKRecvTimeout;
	
public:
	Dispatcher( PIObject *pTheObject, int, const char *[] )
	:
		pObject( pTheObject ),
		pDefaultHost( 0 ),
		bOK( false ),
		iKeepOpen( 1 ),
		pFKRecvTimeout(
			PIDB_getFastKey( KEY_INT_RECVTIMEOUT, PIDBTYPE_OPAQUE ) )
		{
		assert( pObject );
		if ( !pObject )
			{ return; };

		if ( HTTPCore_initServer( pObject ) )
			{
			return;
			};

		/* --- load up default virtual host object --- */
		const char *pDefaultHostName = PIConfig_lookupValue(
			PIObject_getConfigurationDB( pObject ),
			KEY_CONF_DEFAULTHOST, 0, 0 );
		if ( !pDefaultHostName )
			{
			CONFIG_ERR( pObject, "Dispatcher: 'DefaultHost' not defined" );
			return;
			};
		PIDB *pDB = PIObject_getDB( pObject );
		pDefaultHost = PIObject_loadFromLine(
			pDB, PIObject_getConfigurationDB( pObject ), pDefaultHostName );
		if ( !pDefaultHost )
			{/* --- error set by PIObject_loadFromLine() --- */return; };

		/* ---
		Check for connection keep open, optional
		--- */
		const char *pTmp = PIConfig_lookupValue(
			PIObject_getConfigurationDB( pObject ),
			KEY_CONF_KEEPOPEN, 0, 0 );
		if ( pTmp && !PIUtil_stricmp( pTmp, VALUE_OFF ) )
			{ iKeepOpen = 0; };

		bOK = true;
		};

	~Dispatcher()
		{
		PIObject_delete( pDefaultHost, 0, 0 );
		HTTPCore_cleanupServer();
		};

	/* ---
	Invoke all the steps on all the handlers
	--- */
	int HandleConnection( PIHTTP &tPIHTTP )
		{
		int iRet = HTTPCore_dispatch( &tPIHTTP, PH_INIT, PH_DESTROY );
		return iRet;
		};

	/* ---
	Invoke all the steps on all the handlers
	--- */
	inline int Execute( int iArgc, const char *ppArgv[] )
		{
		if ( iArgc<1 )
			{ return PIAPI_ERROR; };
		PIObject *pIO = (PIObject *)*ppArgv;
		if ( !pIO )
			{ return PIAPI_ERROR; };

		/* ---
		Get the DB from the IO object to monitor the KeepOpen flag
		--- */
		PIDB *pC = PIObject_getDB( pIO );
		PIDB_add( pC, PIDBTYPE_OPAQUE, KEY_INT_KEEPOPEN, (void *)iKeepOpen, 0 );
		int iDoKeepOpen = 0;
		int iKeepOpenCount = 10;	/* ---
									silently stop more than 10 connection
									reuses
									--- */
		int iRet;
		PIHTTP tPIHTTP( pIO, PIObject_getDB( pDefaultHost ) );
		for(;;)
			{
			/* ---
			Disallow more than the specified number of connection reuses
			--- */
			if ( !iKeepOpenCount )
				{ PIDB_replace( pC, PIDBTYPE_OPAQUE, KEY_INT_KEEPOPEN, 0, 0 ); }
			else
				{ iKeepOpenCount--; };

			/* --- get ready for new request --- */
			tPIHTTP.NewRequest();

			iRet = HandleConnection( tPIHTTP );
		
			/* --- make sure all data was flushed --- */
			tPIHTTP.GetIOBuffer().Flush();
			if ( iRet==PIAPI_COMPLETED )
				{
				/* ---
				Look at what the headers told the browser would happen
				--- */
				iDoKeepOpen = (int)PIDB_lookup( pC, PIDBTYPE_OPAQUE,
					KEY_INT_KEEPOPEN, 0 );
				}
			else
				{
				return iRet;
				};

			/* --- close connection ? --- */
			if ( !iDoKeepOpen )
				{
				return iRet;
				};

			/* ---
			Get the current timeout and set it low for poll on next
			request
			--- */
			assert( pFKRecvTimeout );
			int iTimeout = (int)PIDB_lookup( pC, PIDBTYPE_OPAQUE,
				pFKRecvTimeout, PIDBFLAG_FASTKEY );
			/* --- 
			NOTE: This keep open timeout should be moved to the
			configuration
			--- */
			PIDB_replace( pC, PIDBTYPE_OPAQUE, pFKRecvTimeout,
				(void *)5, PIDBFLAG_FASTKEY );

			/* --- poll for new data --- */
			if ( tPIHTTP.GetIOBuffer().PollBeforeRead()<=0 )
				{ 
				/* --- record error conditions? --- */
				iRet = PIAPI_COMPLETED;
				break;
				};

			/* ---
			Reset the recv timeout
			--- */
			PIDB_replace( pC, PIDBTYPE_OPAQUE, pFKRecvTimeout,
				(void *)iTimeout, PIDBFLAG_FASTKEY );
			};
		return iRet;		/* ---
							NOTE: there are other return points in this fn.
							--- */
		};

	inline int IsOK() 		{ return bOK; };
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int Dispatcher_execute( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	if ( !pObj ) return PIAPI_ERROR;
	return ((Dispatcher *)PIObject_getUserData(pObj))->Execute( iArgc,
		ppArgv );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int Dispatcher_onClassLoad( PIClass_LoadAction eLoad, int,
	const char *[] )
{
	switch( eLoad )
		{
		case STARTUP:
		default:;
		};
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int Dispatcher_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	Dispatcher *pDisp = new Dispatcher( pObj, iArgc, ppArgv );
	if (!( pDisp && pDisp->IsOK() ))
		{
		delete pDisp;
		return PIAPI_ERROR;
		};
	if ( PIObject_setUserData( pObj, pDisp ) )
		{
		delete pDisp;
		CONFIG_ERR( pObj, "Dispatcher_constructor: \
PIObject_setUserData() failed" );
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int Dispatcher_copyConstructor( PIObject *, int, const char *[] )
{
	return PIAPI_ERROR;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int Dispatcher_destructor( PIObject *pObj, int,
	const char *[] )
{
	delete (Dispatcher *)PIObject_getUserData( pObj );
	return PIAPI_COMPLETED;
}

