/*___________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	(http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *___________________________________________________________________________*|
 *
 * $Source: KrnlSync.cpp$
 * $Date: Sun Aug 10 06:29:02 1997$
 *
 Description:
	Implementation of functions when no threading is supported or
	kernel threads are used.

\*___________________________________________________________________________*/
//$SourceTop:$

#include "OSConfig.h"		/* --- get build options --- */

/* ---
 Only use this file for non user context threads.
--- */
#if !defined(CONFIG_MT_USER)

#include <errno.h>

#if !defined(CONFIG_CPP_EXCEPTIONS)
#	include <setjmp.h>
#endif

/*
** include some windows stuff
*/
#if defined(CONFIG_USE_WINSOCK)
extern "C"
	{
	#	include <winsock.h>					/* for timeval, fd_set etc. */
	};
#endif

#if defined(CONFIG_OS_POSIX)
#	include <unistd.h>
#	include <sys/file.h>
#	include <sys/types.h>
#	include <sys/wait.h>

#elif defined(CONFIG_OS_WIN32)
#elif defined(CONFIG_OS_WIN16)
#endif

#include "Allocate.h"
#include "Platform.h"
#include "PlatDefs.h"

/*___________________________________________________________________________*\
 *
 Typedefs:
\*___________________________________________________________________________*/
#if defined(CONFIG_OS_Stub)
	typedef thread_t THREAD_T;
	typedef thread_key_t THREAD_KEY_T;
#	define BAD_THREAD ((THREAD_T)-1)
#elif defined(CONFIG_OS_WIN32)
	typedef HANDLE THREAD_T;
	typedef DWORD THREAD_KEY_T;
#	define BAD_THREAD ((THREAD_T)0)
#endif

/*___________________________________________________________________________*\
 *
 Declarations:
\*___________________________________________________________________________*/
#if !defined(CONFIG_CPP_EXCEPTIONS)
static jmp_buf ___tJmpBuf;
#endif
// #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
#define D

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	This function will only be called after Platform::Internal_InitThreads so
	it is safe to assume the thread key tTlsAllocator has been setup.
\*___________________________________________________________________________*/
Allocator *Platform::GetCurrentAllocator()
{
	return GetGlobalAllocator();
}

#if !defined(CONFIG_MULTITHREADED)
/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void Platform::Internal_EnterThreads( void (* fn)(void) )
{
	Platform::CatchExceptions( (void(*)(void*))fn, 0 );
}

/*
** Stub functions for mutexes, etc.
*/
/*___________________________________________________________________________*\
 *
 Class:
 Description:
\*___________________________________________________________________________*/
class StubLocalMutex : public Semaphore
{
private:
public:
	StubLocalMutex()										{};
	virtual ~StubLocalMutex()								{};
	virtual int Lock()										{ return 0; };
	virtual int TryLock()									{ return 0; };
	virtual int UnLock()									{ return 0; };
	virtual bool IsOK() const								{ return true; };
};

/*___________________________________________________________________________*\
 *
 Class:
 Description:
\*___________________________________________________________________________*/
class StubLocalSemaphore : public Semaphore
{
private:
public:
	StubLocalSemaphore( int iInitialCount, int iMaxCount )	{};
	virtual ~StubLocalSemaphore()							{};
	virtual int Lock()										{ return 0; };
	virtual int TryLock()									{ return 0; };
	virtual int UnLock()									{ return 0; };
	virtual bool IsOK() const								{ return true; };
};

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Semaphore *Platform::AllocLocalMutex()
{
	return PI_NEW( StubLocalMutex() );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Semaphore *Platform::AllocLocalSemaphore( int iInitialCount, int iMaxCount )
{
	return PI_NEW( StubLocalSemaphore( iInitialCount, iMaxCount ) );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Thread *Platform::AllocThread( void *, int )
{
	PIERROR( PIAPI_NOTSUPPORTED );
	return 0;
}
#endif

#if !defined(CONFIG_MULTITHREADED)
/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::YieldThread()
{
#if defined(CONFIG_OS_POSIX)
#	if !defined(CONFIG_NO_YIELD)
		::yield();

#	endif

#elif defined(CONFIG_OS_WIN32)
	::Sleep( 0 );	/* --- This is how Win32 gives up timeslice --- */

#elif defined(CONFIG_OS_WIN16)
	::Yield();

#else
#	error Unsupported!
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::WaitForThread( Thread * )
{
	PIERROR( PIAPI_NOTSUPPORTED );
	return -1;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Thread *Platform::GetCurrentThread()
{
	PIERROR( PIAPI_NOTSUPPORTED );
	return 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Thread *Platform::GetMainThread()
{
	PIERROR( PIAPI_NOTSUPPORTED );
	return 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::AllocThreadKey()
{
	PIERROR( PIAPI_NOTSUPPORTED );
	return 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
bool Platform::DeleteThreadKey( int )
{
	PIERROR( PIAPI_NOTSUPPORTED );
	return 0;
}
#endif

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::ThreadDbgDump( ostream &os )
{
#if defined(CONFIG_MULTITHREADED)
#	if defined(CONFIG_OS_POSIX)
		os << "Stub threads: no information available" << endl;
#	elif defined(CONFIG_OS_WIN32)
		os << "Win32 threads: no information available" << endl;
#	endif
#else
	os << "No threads supported" << endl;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::PollFD( int iFD, int iFlags, int iTimeout )
{
#if defined(CONFIG_OS_POSIX)
#   if defined(CONFIG_OS_SOLARIS)
		/*
		** Solaris 2.x: man -s 3C select
		**
		** [File descriptors associated with regular files always select
		** true  for  ready  to  read, ready to write, and error conditions.]
		**
		** NOTE:
		**	  - What about pipes?
		**	  - Also change other instances of this function
		*/
		/* so just return the requested status */
		return iFlags;
#   endif
	return Platform::PollNetFD( iFD, iFlags, iTimeout );
#elif defined(CONFIG_OS_WIN32)
	/* ---
	Win32 SDK, says that WaitForSingleObject() can be used to
	check if a file handle operation will succeed without blocking
	'under certain circumstances' and is 'discouraged'. But there doesn't
	seem to be another way to do this without side effects.
	NOTE: The reason this is discouraged is because confusion can arise if
	multiple waits exists for the the same handle, therefore its
	suggested to use an event. Anyway.....
	--- */
	int iRetFlags = 0;
	switch( ::WaitForSingleObject( (HANDLE)iFD, iTimeout*1000 ) )
		{
		case WAIT_FAILED:
			PIOSERR;
			return -1;

        case WAIT_TIMEOUT:
            return 0;

        case WAIT_OBJECT_0:
			/* ---
			Don't know what made the handle signalled so set the
			return code based on what the callee requested. Files
			are unidirectional anyway so this doesn't cause confusion
			like it would with sockets
			--- */
			if ( iFlags & Platform::POLL_READ )
				{ iRetFlags = iRetFlags | Platform::POLL_READ; };
			if ( iFlags & Platform::POLL_WRITE )
				{ iRetFlags = iRetFlags | Platform::POLL_WRITE; };
			return iRetFlags;

		default:
            assert( 0 );        /* what's this ? */
			PIOSERR;
			return -1;
		};
#elif defined(CONFIG_OS_WIN16)
	/*
	** Return ready without yielding
	*/
	(void)iFD;
	(void)iTimeout;
	return iFlags;

#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::PollPipeFD( int iFD, int iFlags, int iTimeout )
{
#if defined(CONFIG_OS_WIN32)
	int iRetFlags = 0;

	if ( iFlags & ~Platform::POLL_WRITE )
		{
		int iRet = Platform::PollFD( iFD, iFlags & ~Platform::POLL_WRITE,
			iTimeout );

		if ( iRet<0 )
			{ return iRet; };

		iRetFlags |= iRet;
		};

	/* ---
	Named pipes on Win32. Always return 'Ready to Write'.
	--- */
	if ( iFlags & Platform::POLL_WRITE )
		{ iRetFlags |= Platform::POLL_WRITE; };

	return iRetFlags;
#endif
	/*
	** Otherwise behaviour is the same as Platform::PollFD
	*/
	return Platform::PollFD( iFD, iFlags, iTimeout );	
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::PollNetFD( int iFD, int iFlags, int iTimeout )
{
	if ( iFD<0 ) 
		{ PIERROR( PIAPI_EINVAL ); return -1; };	
	timeval tv;
	tv.tv_sec = iTimeout;
	tv.tv_usec = 0;
	fd_set tRead, tWrite, tExcept;
	FD_ZERO( &tRead );
	FD_ZERO( &tWrite );
	FD_ZERO( &tExcept );
	if ( iFlags & Platform::POLL_READ )
		{ FD_SET( iFD, &tRead ); };	
	if ( iFlags & Platform::POLL_WRITE )
		{ FD_SET( iFD, &tWrite ); };	
	FD_SET( iFD, &tExcept );	
	int iRet = 0;
	do	{
		iRet = ::select( iFD+1, &tRead, &tWrite, &tExcept,
			iTimeout==-1 ? 0 : &tv );
		} while( iRet==-1 && errno==EINTR );
	switch( iRet )
		{
		case -1:	/* error */
			PIOSERR;
			return -1;

		case 0:		/* timeout */
			return 0;

		default:;	/* num descriptors (exact meaning varies by OS) */
		};
	int iRetFlags = 0;
	if ( FD_ISSET( iFD, &tExcept ) )	{ PIERROR( PIAPI_ERROR ); return -1; };
	if ( FD_ISSET( iFD, &tRead ) )
		{ iRetFlags = iRetFlags | Platform::POLL_READ; };
	if ( FD_ISSET( iFD, &tWrite ) )
		{ iRetFlags = iRetFlags | Platform::POLL_WRITE; };
	return iRetFlags;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::Sleep( int iMilliseconds )
{
#if defined(CONFIG_OS_POSIX)
	::sleep( (iMilliseconds/1000)+((iMilliseconds%1000)?1:0) );
#elif defined(CONFIG_OS_WIN32)
	::Sleep( iMilliseconds );
#elif defined(CONFIG_OS_WIN16)
	(void)iMilliseconds;
	/* LATER what to do here? */
#endif
	return 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::CatchExceptions( void (* fn)( void *), void *pArg )
{
#if defined(CONFIG_CPP_EXCEPTIONS)
	try	{
		(fn)( pArg );
		}
	catch( StubThread::StackUnwindException *pE )
		{
		PI_DELETE( pE );
		}
#else
	if ( setjmp( ___tJmpBuf )==0 )
		{ (fn)( pArg ); };
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::Write( int iFd, int iLength, const void *pData )
{
	assert( iFd!=-1 && iLength && pData );
	if ( iFd==-1 || !iLength || !pData )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};
	int iToWrite = iLength;
	int iWritten;
	for(; iToWrite ;)
		{
		iWritten = -1;
#if defined(CONFIG_OS_POSIX)
		iWritten = ::write( iFd, pData, iLength );
		if ( iWritten==-1 && errno==EINTR )
			{ continue; };

#elif defined(CONFIG_OS_WIN32)
		/* ---
		the cast from int * to DWORD * (DWORD is currrently unsigned long), is
		not	so bad because iLength is the maximum size that can be sent and 
		is also an int
		--- */
		DWORD dwWritten;
		if ( !::WriteFile( (HANDLE)iFd, pData, iLength, &dwWritten, 0 ) )
			{
			dwWritten = (DWORD)-1;
			}
		else
			{
			iWritten = (int)dwWritten;
			};

#endif
		if ( !iWritten || iWritten==-1 )
			{ break; };
		iToWrite -= iWritten;
		};

	if ( iWritten==-1 )
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::WriteAtomic( int iFd, int iLength, const void *pData )
{
	assert( iFd!=-1 && iLength && pData );
	if ( iFd==-1 || !iLength || !pData )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};
	int iToWrite = iLength;
	int iWritten;
#if defined(CONFIG_OS_POSIX)
	/* ---
	Enter loop to obtain a lock on the file
	--- */
	for(;;)
		{
		/* --- blocking call to lock --- */
#if defined(CONFIG_LOCK_FLOCK)
		if ( ::flock( iFd, LOCK_EX )==0 )
#else
		if ( ::lockf( iFd, F_LOCK, 0 )==0 )
#endif
			{
			/* --- got the lock, continue with accept() --- */
			break;
			}
		else if ( errno==EINTR )
			{
			/* ---
			man on solaris did not list this as a possible scenario,
			i.e. operation interrupted because a signal was caught,
			but just in case...
			--- */
			continue;
			};

		/* --- an error occurred in lock --- */
		PIOSERR;
		return PIAPI_ERROR;
		};
#endif

	for(; iToWrite ;)
		{
		iWritten = -1;

#if defined(CONFIG_OS_POSIX)
		iWritten = ::write( iFd, pData, iLength );
		if ( iWritten==-1 && errno==EINTR )
			{ continue; };

#elif defined(CONFIG_OS_WIN32)
		/* --- 
		the cast from int * to DWORD * (DWORD is currrently unsigned long), is
		not	so bad because iLength is the maximum size that can be sent and 
		is also an int
		--- */
		if ( !::WriteFile( (HANDLE)iFd, pData, iLength, (DWORD *)&iWritten,
			0 ) )
			{ iWritten=-1; };

#endif
		if ( !iWritten || iWritten==-1 )
			{ break; };
		iToWrite -= iWritten;
		};

#if defined(CONFIG_OS_POSIX)
	/* ---
	Release the lock
	--- */
#if defined(CONFIG_LOCK_FLOCK)
	if ( ::flock( iFd, LOCK_UN )==-1 )
#else
	if ( ::lockf( iFd, F_ULOCK, 0 )== -1)
#endif
		{ iWritten = -1; };
	
#endif

	if ( iWritten==-1 )
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::WaitForProcess( int tProcess, int iFlags, int iTimeout )
{
	(void)iFlags;	/* not used */
	(void)iTimeout; /* not used */
#if defined(CONFIG_OS_POSIX)
	int iRet = ::waitpid( tProcess, 0, 0 );
	if ( iRet==-1 && errno!=EINTR && errno!=ERESTART )
		{
		PIOSERR;
		return PIAPI_ERROR;
		};

	if ( iRet==tProcess )
		{ return PIAPI_COMPLETED; };

	/* --- how'd we get here? --- */
	assert( 0 );
	PIERROR( PIAPI_ERROR );
	return PIAPI_ERROR;

#elif defined(CONFIG_OS_WIN32)
	DWORD dwTimeout = ( iTimeout==-1 ) ? INFINITE : iTimeout*1000 ;
	DWORD dwWait=::WaitForSingleObject( (HANDLE)tProcess, dwTimeout );
	switch( dwWait )
		{
		case WAIT_OBJECT_0:
			return PIAPI_COMPLETED;

		case WAIT_FAILED:
			PIOSERR;
			return PIAPI_ERROR;

		case WAIT_TIMEOUT:
			PIERROR( PIAPI_TIMEOUT );
			return PIAPI_TIMEOUT;

		default:
			/* how did we get here */
			assert( 0 );
			PIERROR(PIAPI_ERROR);
			return PIAPI_ERROR;
		};
#endif
	(void)tProcess;
	return PIAPI_NOTSUPPORTED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::LockFd( int iFd )
{
	assert( iFd!=-1 );
	if ( iFd==-1 )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};

#if defined(CONFIG_OS_POSIX)
	/* ---
	Enter loop to obtain a lock on the file descriptor
	--- */
	for(;;)
		{
		/* --- blocking call to lock --- */
#if defined(CONFIG_LOCK_FLOCK)
		if ( ::flock( iFd, LOCK_EX )==0 )
#else
		if ( ::lockf( iFd, F_LOCK, 0 )==0 )
#endif
			{
			/* --- got the lock, continue with accept() --- */
			break;
			}
		else if ( errno==EINTR )
			{
			/* ---
			man on solaris did not list this as a possible scenario,
			i.e. operation interrupted because a signal was caught,
			but just in case...
			--- */
			continue;
			};

		/* --- an error occurred in lock --- */
		PIOSERR;
		return PIAPI_ERROR;
		};
#endif

	return PIAPI_COMPLETED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::UnlockFd( int iFd )
{
	assert( iFd!=-1 );
	if ( iFd==-1 )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};

	/* ---
	Release the lock
	--- */
#if defined(CONFIG_OS_POSIX)

#if defined(CONFIG_LOCK_FLOCK)
	if ( ::flock( iFd, LOCK_UN )==-1 )
#else
	if ( ::lockf( iFd, F_ULOCK, 0 )== -1)
#endif
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
#endif

	return PIAPI_COMPLETED;
}

#endif /* !defined(CONFIG_MT_USER) */

