/*___________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE       
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *___________________________________________________________________________*|
 *
 * $Source: UserSync.cpp$
 * $Date: Sun Aug 10 06:29:10 1997$
 *
 Description:
	Implementation of user context threads and synchronization 
	facilities.

\*___________________________________________________________________________*/
//$SourceTop:$

#include "OSConfig.h"		/* --- get build options --- */
#include "PlatDefs.h"

/* ---
 Include this file only for builds which use user multithreading
--- */
#if defined(CONFIG_MULTITHREADED) && defined(CONFIG_MT_USER)

/*___________________________________________________________________________*\
 *
 Description:
	Debugging macros
\*___________________________________________________________________________*/
// #define DEBUG	40		/* >=40  ugly, verbose and line numbers */
// #define DEBUG	30		/* >=30 ugly verbose runtime logging */
// #define DEBUG	25		/* >=25 more verbose runtime logging */
// #define DEBUG	20		/* >=20 verbose runtime logging */
// #define DEBUG	15		/* >=15 terse runtime logging */
#define DEBUG	10		/* >=10 verbose logging on catastrophe */
// #define DEBUG	5		/* >0	terse logging on catastrophe */
// #define DEBUG	0		/* ==0 never log */

#if DEBUG>=40
#	define D { CERR << "[" << Platform::GetProcessId() << ":" << \
		(___pTheCurrentThread ? (int)___pTheCurrentThread->GetSystemHandle() \
		: -1 ) << "] " << __FILE__ << \
		": " << __LINE__ << endl; }
#else
#	define D
#endif

#if !defined(CONFIG_CPP_EXCEPTIONS)
#	include <setjmp.h>
#endif

#if defined(CONFIG_OS_POSIX)
	#	include <unistd.h>
	#	include <sys/types.h>
	#	include <sys/time.h>
	#	include <sys/wait.h>
	#	if defined(CONFIG_USE_BSTRING)
	#		include <bstring.h>				/* bzero() used by FD_ZERO */
	#	endif
	# 	if defined(CONFIG_LOCK_FLOCK)
	#		define _BSD_COMPAT			  	/* conflict on SGI, see man flock */
	#		include <sys/file.h>			/* for ::flock() */
	#	endif
#elif defined(CONFIG_OS_WIN16)
	#	if defined(__BORLANDC__)
		/* don't want those borland 'cant inline' messages */
	#		pragma warn -inl
	#	endif
	#
#endif

/*
** include some windows stuff
*/
#if defined(CONFIG_USE_WINSOCK)
extern "C"
	{
	#	include <winsock.h>					/* for timeval, fd_set etc. */
	};
#endif

#include <iostream.h>
#include <time.h>
#include <limits.h>
#include <errno.h>

#include "UserSync.h"
#include "PICompat.h"
#include "Allocate.h"
#include "OSEntity.h"
#include "Platform.h"

#if 0
// LATER
#include <fstream.h>
ofstream ofs( "D:\\DEVEL\\Pi3WEB\\SOURCE\\out.lst" );
#define CERR ofs
#endif
#define CERR cerr

/*___________________________________________________________________________*\
 *
 Declarations:
\*___________________________________________________________________________*/
/*
** Stack sizes
**
**		0x04000		=		16K
** 		0x08000		=		32K
**		0x10000		=		64K
**		0x18000		=		96K
**		0x20000		=	   128K
**
*/
#if defined(CONFIG_OS_WIN16)
	/*
	** Small stacksize
	*/
#	define PI_DEFAULTSTACKSIZE	0x2000
#else
	/*
	** Flat memory models
	*/
#	define PI_DEFAULTSTACKSIZE	0x10000
#endif

#define PI_MAXTHREADKEYS		128			/* Maximum number of keys */

/* ---
size of register context measured in units of size_t size
This is architecure dependent.
--- */
#if defined(CONFIG_ASM_I386)
#	define PI_CONTEXTSIZE		8
#	define PI_STACKDISPLACEMENT	0x100
#elif defined(CONFIG_ASM_SPARC)
#	define PI_CONTEXTSIZE		4
#	define PI_STACKDISPLACEMENT	0x100
#elif defined(CONFIG_ASM_MIPS)
#	define PI_CONTEXTSIZE		11
#	define PI_STACKDISPLACEMENT	0x100
#elif defined(CONFIG_ASM_HP)
#	define PI_CONTEXTSIZE		4
#	define PI_STACKDISPLACEMENT	0x100
#else
#	error Unsupported Machine Architecture
#endif

class UserLocalSemaphore;
class UserThread;
class ThreadScheduler;
class PollInfo;
class Poller;

int aUsedKeys[PI_MAXTHREADKEYS];			/* must be zero initialized */
static ThreadScheduler *___pLatentThreads = 0;	/* global thread scheduler */
static Poller *___pThePoller = 0;				/* polls on sockets/files */
static UserThread *___pTheMainThread = 0;		/* the main thread */
static UserThread *___pTheCurrentThread = 0;	/* the current thread */
#define ___refTheCurrentThread (UserThread *)___pTheCurrentThread)
static UserThread *___pNewThread = 0;			/* a newly created thread */
static UserThread *___pExitingThread = 0;		/* an exiting thread */
static int iThreadsCreated = 0;					/* since initialization */
extern "C" int StackContextSwitch(void);		/* most interesting function */
void DbgWriteAllThreads( ostream &os );			/* for debugging */

/*___________________________________________________________________________*\
 *
 Class:
 Description:
\*___________________________________________________________________________*/
class ThreadScheduler
{
private:
	DblList lThreads[PITHREAD_PRIORITY_SIZE];
	int iNumThreads;

	/* --- forbid copy constructor --- */
	ThreadScheduler( const ThreadScheduler & )
	:	iNumThreads( 0 )
		{ assert( 0 ); };

public:
	ThreadScheduler()
	:	iNumThreads( 0 )
		{};

	void AddThread( UserThread *pThread );

	/* ---
	Remove a latent thread from the list of threads, assumes
	that there is only one entry per thread
	--- */
	bool RemoveThread( UserThread *pThread )
		{
		for( int i=0; i<PITHREAD_PRIORITY_SIZE; i++)
			{
			DblList &tList = lThreads[i];
			for( DblListIterator j( tList ); !j.BadIndex(); j++)
				{
				if ( pThread == j.Current() )
					{	
					tList.Detach( j );
					iNumThreads--;
					RemoveThread( pThread );
					return true;
					};
				};
			};
		return false;
		}

	UserThread *SelectNextThread();
	UserThread *SelectNextRunnableThread( int iMandatory );
	bool ChangeThreadPriority( UserThread *pThread, int iOld, int iNew );
	void DbgWrite( ostream &os );
	inline int Size() const		{ return iNumThreads; };
};


// LATER
#if defined(CONFIG_OS_WIN16)
static WORD __wLastStackPos = 0;
#endif

/*___________________________________________________________________________*\
 *
 Class:
 Description:
\*___________________________________________________________________________*/
class UserThread : public Thread
{
public:
#if defined(CONFIG_CPP_EXCEPTIONS)
	class StackUnwindException
		{
		};
#else
	jmp_buf tJmpBuf;
#endif

private:
	void *aKeys[PI_MAXTHREADKEYS];	/* thread local data */
	int iThreadId;				/* thread identifier, for description */
	ThreadFn fnEntry;			/* user entry point for thread */
	unsigned long ulData;		/* data to the user entry point function */
	Allocator *pAllocator;		/* allocator for this thread */
	void *pStack;				/* the stack for this thread */
	unsigned long ulStatus;		/* final return code */
	int iPriority;				/* thread priority */
	bool bIsTerminated;			/* thread is dead */
	ThreadScheduler tWaiting;	/* threads waiting for this thread to die */
	int iStackSize;				/* size of stack to allocate */
	int iNextReturnCode;		/* result of a socket/file poll or sync. wait */

	/* --- runnable status (whether thread is blocked, and why) --- */
	UserLocalSemaphore *pSemaphore;		/* a semaphore we are waiting on */
	UserThread *pThread;				/* a thread we are waiting on to die */
	PollInfo *pPoll;					/* polling on socket/file, or asleep */
	bool bSuspended;					/* suspended */

	/* --- the thread context --- */ 
	size_t aContext[PI_CONTEXTSIZE];	/* for saving register context */

	/* --- put this thread to sleep, because of a semaphore --- */
	int PutToSleep( UserLocalSemaphore *pTheSemaphore )
		{
		assert( !pSemaphore );
		pSemaphore = pTheSemaphore;
		StackContextSwitch();
		int iTmp = iNextReturnCode;
		iNextReturnCode = 0;
		return iTmp;
		};

	/* --- put this thread to sleep, because of another thread --- */
	int PutToSleep( UserThread *pTheThread )
		{
		/* --- because we're waiting on another thread to die --- */
		assert( !pThread && pTheThread );
		pThread = pTheThread;
		pThread->tWaiting.AddThread( this );
		iNextReturnCode = 0;
		StackContextSwitch();
		int iTmp = iNextReturnCode;
		iNextReturnCode = 0;
		return iTmp;
		};

	/* --- put this thread to sleep, for no reason (suspend) --- */
	int PutToSleep()
		{
		StackContextSwitch();
		int iTmp = iNextReturnCode;
		iNextReturnCode = 0;
		return iTmp;
		};

	/* --- put this thread to sleep, because of socket poll --- */
	int PutToSleep( int iFD, int iFlags, int iTimeout );

	/* ---
	Force end of sleep, poll or wait, propably because of deadlock
	--- */
	void AbortBlocks( int iWhy );

 	/* --- make this thread runnable (or closer to runnable) --- */
	void MakeRunnable( UserLocalSemaphore *pTheSemaphore )
		{
		assert( pTheSemaphore==pSemaphore );
		pSemaphore = 0;
		};

	/* --- make this thread runnable (or closer to runnable) --- */
	void MakeRunnable( UserThread *pTheThread )
		{
		assert( pThread == pTheThread );
		pThread = 0;
		};

	/* --- make this thread runnable (or closer to...) --- */
	inline void MakeRunnable()
		{
		bSuspended = false;
		};

	/* --- make this thread runnable (or closer to runnable) --- */
	void MakeRunnable( int iThePollResult );

	void Cleanup();

	/* ---
	notification to this thread that a waiting thread is no longer
	waiting
	--- */
	void GaveUpWaiting( UserThread *pTheThread )
		{
		if ( !tWaiting.RemoveThread( pTheThread ) )
			{
			assert( 0 );		/* why did that thread notify us? */
			};
		};

	bool InternalSetPriority( int iThePriority );
	bool InternalTerminate( unsigned long ulTheStatus );

	/* --- forbid copy constructor --- */
	UserThread( const UserThread & )
	:	iThreadId( 0 ),
		fnEntry( 0 ),
		ulData( 0 ),
		pAllocator( 0 ),
		pStack( 0 ),
		ulStatus( 0UL ),	
		iPriority( PITHREAD_PRIORITY_MED ),
		bIsTerminated( true ),
		iStackSize( 0 ),		/* main thread has stack size=0 */
		iNextReturnCode( 0 ),
		pSemaphore( 0 ),	
		pThread( 0 ),
		pPoll( 0 ),
		bSuspended( false )
		{ assert( 0 ); };

public:
	UserThread( 
			Allocator *pTheAllocator = Platform::GetGlobalAllocator(),
			int iTheStackSize = PI_DEFAULTSTACKSIZE )
	:	iThreadId( iThreadsCreated++ ),
		fnEntry( 0 ),
		ulData( 0 ),
		pAllocator( pTheAllocator ),
		pStack( 0 ),
		ulStatus( 0UL ),	
		iPriority( PITHREAD_PRIORITY_MED ),
		bIsTerminated( true ),
		iStackSize( iTheStackSize ),	/* main thread has stack size=0 */
		iNextReturnCode( 0 ),
		pSemaphore( 0 ),	
		pThread( 0 ),
		pPoll( 0 ),
		bSuspended( false )
		{
#if DEBUG>=20
		CERR << "User Thread constructor: " << (void *)this << endl;
#endif
		if ( !pAllocator )
			{ return; };
		memset( aKeys, 0, sizeof( int ) * PI_MAXTHREADKEYS );
		if ( iStackSize )
			{
#	if !defined(CONFIG_OS_WIN16)
			pStack = Platform::AllocMemory( iStackSize );
#	else
			/*
			** Win16 - The segmented world!
			**
			** Allocate a stack for this thread from another part
			** of the main stack
			*/
			if ( __wLastStackPos==0 )
				{
				__wLastStackPos=(WORD)__Pi_GetCurrentStackPtr();
				};
			__wLastStackPos -= ( PI_DEFAULTSTACKSIZE );
			pStack = (void *)__wLastStackPos;
#	endif
			};
		};

	bool Begin( 
			ThreadFn fnTheEntry,
			unsigned long ulTheData, 
			int iThePriority,
			int iFlags )		
		{
#if DEBUG>=20
		CERR << "User Thread Begin(): " << (void *)this << endl;
		DbgWrite( CERR );
		CERR << endl;
#endif
		if ( !bIsTerminated )
			{
			PIERROR( PIAPI_EINVAL );
			return false;
			};
		if ( !fnTheEntry )
			{
			PIERROR( PIAPI_EINVAL );
			return false;
			};
		if ( iFlags & PITHREAD_FLAGS_SUSPENDED )
			{ bSuspended = true; };

		if ( this!=___pTheMainThread && !pStack )
			{
			PIERROR( PIAPI_EINVAL );
			return false;
			};
		/* --- reset these values --- */
		ulStatus = 0UL;
		bIsTerminated = false;
		fnEntry = fnTheEntry;
		ulData = ulTheData;
		InternalSetPriority( iThePriority );
		if ( pStack )
			{
			/* --- this is a regular thread --- */
			/* ---
			Set initial stack pointer assuming that the stack
			grows downwards
			--- */
#if defined(CONFIG_ASM_I386)
#	if defined(CONFIG_ASM_I386_WIN16)
			/* --- win 16 stack pointer is already setup */
			aContext[0] = ((size_t)pStack);
#	else
			aContext[0] = ((size_t)pStack) + ((size_t)iStackSize)
				- PI_STACKDISPLACEMENT;	// down
#	endif
#elif defined(CONFIG_ASM_SPARC)
			aContext[0] = ((size_t)pStack) + ((size_t)iStackSize)
				- PI_STACKDISPLACEMENT;	// down
#elif defined(CONFIG_ASM_MIPS)
			aContext[0] = ((size_t)pStack) + ((size_t)iStackSize)
				- PI_STACKDISPLACEMENT;	// down
#elif defined(CONFIG_ASM_HP)
			aContext[0] = ((size_t)pStack) + ((size_t)iStackSize)
				- PI_STACKDISPLACEMENT;	// down
#else
#			error Unsupported Machine Architecure

			/* ---
			If the stack grew up this would be
			--- */
			aContext[0] = ((size_t)pStack);				// up
#endif
			___pNewThread = this;
			StackContextSwitch();		/* switch context to new thread */
			}
		else
			{
			/* --- this is the main thread --- */
			BeginThread();	
			};
		return true;
		};

	virtual ~UserThread()
		{
#if DEBUG>=20
		CERR << "User Thread destructor: " << (void *)this << endl;
#endif
		if(  !pStack )
			{
			/* --- verify this is the running thread, because this destructor
			must have been invoked from the thread cleanup routine
			--- */
			PIMUSTBETRUE( ___pTheCurrentThread == this );
			return;
			};
		if ( !bIsTerminated )
			{ InternalTerminate( (unsigned long)-1 ); };
		WaitForThread( this );
#if !defined(CONFIG_OS_WIN16)
		bool bRet = Platform::FreeMemory( pStack );
		assert( bRet );
#endif
		};

	virtual bool Suspend()
		{
		if ( bSuspended )
			{ return true; };
		bSuspended = true;
	
		/* --- if some this is not the current thread just mark it --- */
		if ( this != ___pTheCurrentThread )
			{ return true; };

		/* --- this is the current thread, yield into suspension --- */
		int iRet = PutToSleep();
		if ( iRet )
			{ PIERROR(iRet); return false; }; 
		return true;
		};

	virtual bool Resume()
		{
		if ( !bSuspended )
			{ return true; };
		assert( this != ___pTheCurrentThread );	/* 'coz that would be wierd */
		MakeRunnable();
		return true;
		};

	virtual int GetPriority()
		{ return iPriority; };

	virtual bool SetPriority( int iThePriority );

	virtual bool SetData( int iKey, void *pData )
		{
		if ( iKey<0 || iKey>=PI_MAXTHREADKEYS || 
			!aUsedKeys[ iKey ] )
			{ PIERROR( PIAPI_EINVAL ); return false; };
		aKeys[iKey] = pData;
		return true;
		};

	virtual bool GetData( int iKey, void **ppData )
		{
		if ( iKey<0 || iKey>=PI_MAXTHREADKEYS ||
			!aUsedKeys[ iKey ] || !ppData )
			{ PIERROR( PIAPI_EINVAL ); return false; };
		*ppData = aKeys[iKey];
		return true;
		};
	
	virtual bool Terminate( unsigned long ulTheStatus )
		{
		if ( bIsTerminated )
			{ PIERROR( PIAPI_EINVAL ); return false; };
		return InternalTerminate( ulTheStatus );
		};

	virtual void *GetSystemHandle()
		{ return (void *)iThreadId; };

	/* --- new functions --- */
	inline size_t *GetContext()		{ return aContext; };
	inline bool IsRunnable()
		{ return !pSemaphore && !bSuspended && !pThread && !pPoll; };

	/* --- utility member functions --- */
	void StackUnwind()
		{
#if defined(CONFIG_CPP_EXCEPTIONS)
		throw PI_NEW( StackUnwindException() );
#else
		longjmp( tJmpBuf, 73 );
#endif
		};

	void BeginThread()
		{
#if defined(CONFIG_CPP_EXCEPTIONS)	
		try {
			/* ---
			This StackContextSwitch() is not just important for
			assuming correct priority, its also important to let
			the thread adopt the behaviour defined by its run
			state.
			--- */
			StackContextSwitch();	/* now assume correct priority */
			ulStatus = ( fnEntry )( ulData );
			Cleanup();
			}
		catch( StackUnwindException *pE )
			{
			PI_DELETE( pE );
			Cleanup();
			}
		catch(...)
			{
			Cleanup();
			};
#else
		if ( setjmp( tJmpBuf )==0 )
			{
			StackContextSwitch();	/* now assume correct priority */
#if DEBUG>=20
			DbgWrite( CERR );
			CERR << endl;
#endif
			ulStatus = ( fnEntry )( ulData );
#if DEBUG>=20
			DbgWrite( CERR );
			CERR << endl;
#endif
			};
		Cleanup();
#endif
		};

	int WaitForThread( UserThread *pTheThread )
		{
		if ( !pTheThread->bIsTerminated )
			{ return PutToSleep( pTheThread ); };
		return 0;
		};

	inline bool IsSuspended()		{ return bSuspended; };
	
	void DbgWrite( ostream &os )
		{
		os.width( 4 );
		os << iThreadId << ":" << iPriority << " ";
		os.width( 10 );
#if defined(CONFIG_OS_WIN16)
		os << (void *)this << " SP: ";
		os.width( 12 );
		os << (void *)aContext[0];
		os << " BP: ";
		os.width( 12 );
		os << (void *)aContext[2];
#else
		os << (void *)this << " STACK: ";
		os.width( 12 );
		os << (void *)aContext[0];
#endif
		os << " T(" << tWaiting.Size() << ")\t";
		if ( bIsTerminated )
			{ os << "DEAD"; }
		else if ( IsRunnable() )
			{ os << "CAN RUN"; }
		else
			{
			if ( pSemaphore )
				{ os << "BLOCKED S(" << (void *)pSemaphore << ")"; }
			else if ( pThread )
				{ os << "BLOCKED T(" << (void *)pThread << ")"; }
			else if ( pPoll )
				{ os << "BLOCKED P(" << (void *)pPoll << ")"; }
			else if ( bSuspended )
				{ os << "SUSPENDED"; }
			else
				{ os << "** UNKNOWN **"; };
			};
		};			

	/* --- friends --- */
	friend class UserLocalSemaphore;
	friend class ThreadScheduler;
	friend class PollInfo;
	friend class Poller;
};

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void ThreadScheduler::AddThread( UserThread *pThread )
{
	assert( pThread );
	assert( pThread->iPriority <= PITHREAD_PRIORITY_HIGHEST );
	assert( pThread->iPriority < PITHREAD_PRIORITY_SIZE );
	lThreads[pThread->iPriority].Append( (DblList::type)pThread );	
	iNumThreads++;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
UserThread *ThreadScheduler::SelectNextThread()
{
	for( int i=PITHREAD_PRIORITY_SIZE-1; i>=0; i--)
		{
		DblList &tList = lThreads[i];
		for( DblListIterator j( tList ); !j.BadIndex(); j++)
			{
			UserThread *pThread = (UserThread *)j.Current();
			assert( pThread );
			tList.Detach( j );
			iNumThreads--;
			return pThread;
			};
		};
	return 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
UserThread *ThreadScheduler::SelectNextRunnableThread( int iMandatory )
{
	/* ---
	Get the lowest priority thread to run. If this is not a mandatory
	switch don't select a thread with lower priority that the current
	thread
	--- */
	int iLowestPriority = iMandatory ? 0 : ___pTheCurrentThread->GetPriority();
	for( int i=PITHREAD_PRIORITY_SIZE-1; i>=iLowestPriority; i--)
		{
		DblList &tList = lThreads[i];
		for( DblListIterator j( tList ); !j.BadIndex(); j++)
			{
			UserThread *pThread = (UserThread *)j.Current();
			assert( pThread );
			if ( pThread->IsRunnable() )
				{
				tList.Detach( j );
				iNumThreads--;
				return pThread;
				};
			};
		};

	if ( !iMandatory )
		{
		/* ---
		It isn't mandatory to find another thread, return 0 to indicate
		that the current thread would not be scheduled out.
		--- */
		return 0;
		};

	/* --- OK debug time, the thread scheduling has deadlocked --- */
#	if DEBUG>=20
		CERR << " ** DEADLOCK **" << endl;
#	endif
#	if DEBUG>=20
		DbgWriteAllThreads( CERR );
#	endif
	
	/* ---
	Force the last running thread (___pTheCurrentThread still points
	to it) to run. Do this by aborting its wait state. The blocking function	
	will return back to the callee with a 'DEADLOCK' error code.
	--- */
	PIMUSTBETRUE(___pTheCurrentThread);
	PIMUSTBETRUE(RemoveThread(___pTheCurrentThread));
	___pTheCurrentThread->AbortBlocks( PIAPI_DEADLOCK );
	return ___pTheCurrentThread;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
bool ThreadScheduler::ChangeThreadPriority(
	UserThread *pThread, int iOld, int iNew )
{
	assert( pThread != ___pTheCurrentThread );
	assert(	iOld <= PITHREAD_PRIORITY_HIGHEST &&
		iOld < PITHREAD_PRIORITY_SIZE &&
		iNew <= PITHREAD_PRIORITY_HIGHEST &&
		iNew < PITHREAD_PRIORITY_SIZE );

	/* --- remove thread from old priority list --- */
	for( DblListIterator i( lThreads[iOld] ); !i.BadIndex(); i++)
		{
		if ( i.Current() == pThread )
			{ lThreads[iOld].Detach( i ); break; };
		};
	/* ---
	If the thread has been promoted then place it as the next available
	thread to run in its new priority.
	If the thread has been demoted then put it at the end of the run list
	in its new priority.
	--- */
	if ( iNew > iOld )
		{ lThreads[iNew].PushBack( (DblList::type)pThread ); }
	else
		{ lThreads[iNew].Append( (DblList::type)pThread ); }
	return true;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void ThreadScheduler::DbgWrite( ostream &os )
{
	for( int i=0; i<PITHREAD_PRIORITY_SIZE; i++ )
		{
		for( DblListIterator j( lThreads[i] ); !j.BadIndex(); j++)
			{
			UserThread *pThread = (UserThread *)j.Current();
			assert( pThread );
			os << Platform::GetProcess() << "\t";
			os << "S ";
			pThread->DbgWrite( os );
			os << endl;
			};
		};
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void DbgWriteAllThreads( ostream &os )
{
	os << Platform::GetProcess() << "\t";
	os << "R ";
	___pTheCurrentThread->DbgWrite( os );	
	os << endl;
	___pLatentThreads->DbgWrite( os );
}

/*___________________________________________________________________________*\
 *
 Class:
 Description:
\*___________________________________________________________________________*/
class UserLocalSemaphore : public Semaphore
{
private:
	bool bError;
	int iCurrentCount;
	int iMaxCount;
	ThreadScheduler tWaiting; /* threads which are waiting on this semaphore */

	/* ---
	Queue the current thread on this semaphore, returns 0 on OK.
	--- */
	int InternalSleepThread()
		{
		UserThread *pThread = ___pTheCurrentThread; 
		tWaiting.AddThread( pThread );
#if DEBUG>=20
		CERR << "Sync: " << (void *)this << " putting thread to sleep: ";
		pThread->DbgWrite( CERR );
		CERR << endl;
#endif
		return pThread->PutToSleep( this );
		};

	/* ---
	Free the next thread waiting on this semaphore.
	--- */
	void InternalWakeUpThread()
		{
		/* ---
		this doesn't necessarily wake up the thread, it grants it a
		semaphore lock and marks it runnable
		--- */
		PIMUSTBETRUE( tWaiting.Size() );
		UserThread *pThread = tWaiting.SelectNextThread();
		assert( !pThread->bIsTerminated );
#if DEBUG>=20
		CERR << "Sync: " << (void *)this << " waking up thread: ";
		pThread->DbgWrite( CERR );
		CERR << endl;
#endif
		pThread->MakeRunnable( this );
		};

	/* ---
	notification to this semaphore that a waiting thread is no longer
	waiting
	--- */
	void GaveUpWaiting( UserThread *pTheThread )
		{
#if DEBUG>=20
		CERR << "Sync: " << (void *)this << " waiting thread is exiting: ";
		pTheThread->DbgWrite( CERR );
		CERR << endl;
#endif
		if ( tWaiting.RemoveThread( pTheThread ) )
			{ return; };
#if DEBUG>=20
		CERR << "Sync: no record of waiting thread (" << pTheThread <<
			")" << endl;
		CERR << "Sync: " << tWaiting.Size() << " waiting threads " << endl;
#endif
		assert( 0 ); 			/* why did that thread notify us? */
		};

	/* --- forbid copy constructor --- */
	UserLocalSemaphore( const UserLocalSemaphore & )
	:	bError( true ),
		iCurrentCount( 0 ),
		iMaxCount( 0 )
		{ assert( 0 ); };

public:
	UserLocalSemaphore( int iTheInitialCount, int iTheMaxCount )
	:	bError( true ),
		iCurrentCount( iTheInitialCount ),
		iMaxCount( iTheMaxCount )
		{
		if (	( iTheInitialCount <= iTheMaxCount ) &&
				( iTheInitialCount >= 0 ) &&
				( iTheMaxCount >= 0 ) )
			{ bError = false; };
		if ( bError )
			{ PIERROR( PIAPI_EINVAL ); };
		};

	virtual ~UserLocalSemaphore()
		{
		assert( !tWaiting.Size() );		/* no threads should be waiting */
		while( tWaiting.Size() )
			{ InternalWakeUpThread(); };
		};

	virtual int Lock()
		{
#if DEBUG>=25
		CERR << "Locking, threads:- (iCurrentCount, tWaiting.Size()) " 
			<< iCurrentCount << ", " << tWaiting.Size() << endl;
		DbgWriteAllThreads( CERR );
#endif
		if ( !iCurrentCount )
			{
			/* --- make current thread wait --- */
			int iRet = InternalSleepThread();
			if ( iRet )
				{
				/* --- thread couldn't wait, maybe deadlock? --- */
				PIERROR(iRet); return -1;
				};
			}
		else
			{
			iCurrentCount--;
			};
		return 0;
		};
	
	virtual int TryLock()
		{
		if ( !iCurrentCount )
			{ 
			/* --- semaphore not obtained --- */
			return PISYNC_EBUSY;
			}
		else
			{
			iCurrentCount--;
			};
		return 0;
		};

	virtual int UnLock()
		{
#if DEBUG>=25
		CERR << "Unlocking, threads:- (iCurrentCount, tWaiting.Size()) "
			<< iCurrentCount << ", " << tWaiting.Size() << endl;
		DbgWriteAllThreads( CERR );
#endif
/*
**	NOTE:
**
**	This entire facility (to track which threads have which semaphores):-
**	is flawed. I believe it should be possible for one thread to release
**	a semaphore acquired by another thread. This is useful when blocking on 
**	sufficient threads in a pool, using a semaphore - but the semaphore
**	is acquired by the dispatcher thread and released by the handler
**	thread. I know Win32 allows this. HMMM
** 
**		/.* --- Remove one instance of this
**		semaphore from the list of semaphores the thread has acquired 
**		--- *./
**		for( DblListIterator i( ___pTheCurrentThread->lSemaphores ); 
**				!i.BadIndex(); i++)
**			{
**			if ( i.Current() == this )
**				{
**				___pTheCurrentThread->lSemaphores.Detach( i );
**				break;			/.* 1st instance only ! *./
**				};
**			};
**
** Note that all other code implementing this facility has been removed from
** this source file and this note is left for posterity only.
*/

		if ( tWaiting.Size() )
			{ InternalWakeUpThread(); }
		else
			{
			iCurrentCount++;
			assert( iCurrentCount<=iMaxCount );
			if ( iCurrentCount>iMaxCount ) 
				{ iCurrentCount--; PIERROR( PIAPI_ERROR ); return -1; };	
			};
		return 0;
		};

	virtual bool IsOK() const
		{
		return !bError;
		};

	/* --- friends --- */
	friend class UserThread;
};

/*___________________________________________________________________________*\
 *
 Class:
 Description:
\*___________________________________________________________________________*/
class PollInfo
{
private:
	timeval tvTimeout;
	Timer *pTimer;
	int iFlags;
	int iFD;

	void Sync()
		{
		long lSec, lUSec;
		PIMUSTBETRUE( pTimer->Elapsed( &lSec, &lUSec ) );	
		pTimer->Reset();
		tvTimeout.tv_sec -= lSec;
		tvTimeout.tv_usec -= lUSec;
		if ( tvTimeout.tv_usec < 0 )
			{
			tvTimeout.tv_sec--;
#if ( UINT_MAX > 0xffff )	/* subtilties */
			tvTimeout.tv_usec += 1000000;
#else
			tvTimeout.tv_usec += 1000000L;
#endif
			};
#if DEBUG>=30
		CERR << "PollInfo::Sync(): fd=" << iFD;
		if ( iFlags & Platform::POLL_READ ) { CERR << " READ"; };
		if ( iFlags & Platform::POLL_WRITE ) { CERR << " WRITE"; };
		CERR << ", timeout: ";
		if ( !pTimer )
			{ CERR << "never"; }
		else
			{
			CERR << "in " << tvTimeout.tv_sec << ".";
			CERR.width( 6 );
			CERR.fill( '0' );
			CERR << tvTimeout.tv_usec << " seconds";
			CERR.fill( ' ' );
			};
		CERR << endl;
#endif
		};

public:
	PollInfo( int iTheFD, int iTheFlags, int iTheTimeout )
	:	pTimer( iTheTimeout==-1 ? 0 : Platform::AllocTimer() ),
		iFlags( iTheFlags ),
		iFD( iTheFD )
		{
		tvTimeout.tv_sec = iTheTimeout / 1000;
		tvTimeout.tv_usec = iTheTimeout % 1000;
		};

	~PollInfo()
		{
		PI_DELETE( pTimer );
		};

	/* ---
	Determine whether or not this poll object has a timeout associated
	with it
	--- */
	inline bool HasTimeout()
		{
		return pTimer ? true : false;
		};

	/* ---
	If tv is greater than the current time remaining, lower tv to
	be equal to the current time remaining
	Return false if this poll object has timed out.
	If pTv is not null then calculate the minimum time between pTv and
	the time remaining for this object, set pTv to equal this miniumum.
	This method should only be invoked when there is a valid timeout
	(and hence pTimer points to a timeout object). A callee can use
	the method HasTimeout to determine whether or not a timeout exists.
	--- */
	inline bool MinTimeout( timeval *pTv )
		{
		assert( pTimer );
		if ( !pTimer )	{ return true; };
		Sync();
#if DEBUG>=20
		CERR << "Socket: " << iFD << ", time remaining "
			<< tvTimeout.tv_sec << ".";
		CERR.width( 6 );
		CERR.fill( '0' );
		CERR << tvTimeout.tv_usec;
		CERR.fill( ' ' );
		CERR << endl;
#endif
		if ( tvTimeout.tv_sec < 0 )
			{ return false; };
		if ( pTv )
			{
			if ( tvTimeout.tv_sec > pTv->tv_sec )
				{ ; }
			else if ( tvTimeout.tv_sec < pTv->tv_sec )
				{
				pTv->tv_sec = tvTimeout.tv_sec;
				pTv->tv_usec = tvTimeout.tv_usec;
				}
			else if ( tvTimeout.tv_usec < pTv->tv_usec )
				{
				pTv->tv_usec = tvTimeout.tv_usec;
				};
			};
		return true;
		};

	static int Poll( UserThread *pThread, int iFD, int iFlags, int iTimeout );

	/* --- friends --- */
	friend class Poller;
};

/*___________________________________________________________________________*\
 *
 Class:
 Description:
	A Poller polls IO channels on behalf of a number of threads. It
	has its own dedicated thread which does nothing but poll on IO
	channels and mark threads as runnable when a threads IO channel
	changes from a blocking state.
\*___________________________________________________________________________*/
class Poller
{
private:
	UserThread *pPollerThread;
	DblList tPollingThreads;
	static unsigned long PollLoop( unsigned long ulData )
		{
		for(;;)
			{
#if DEBUG>=20
			CERR << " *** Polling Thread *** " << endl;
			DbgWriteAllThreads( CERR );
#endif
			Poller *pPoller = (Poller *)ulData;
			if ( pPoller && (pPoller->tPollingThreads.Size()>0) )
				{
				if ( !pPoller->DoPoll() )
					{
					return (unsigned long)-1;
					};
				}
			else
				{
				assert( pPoller->pPollerThread );
				assert( pPoller->pPollerThread==___pTheCurrentThread );
				assert( !pPoller->pPollerThread->IsSuspended() );
				PIMUSTBETRUE( pPoller->pPollerThread->Suspend() );
				};
			StackContextSwitch();

			/* ---
			StackContextSwitch() can't stall or it will throw the
			program into an infinite loop. Deadlock is the only thing
			that could stall the program.
			--- */
			PIMUSTBETRUE( !pPoller->pPollerThread->iNextReturnCode );
			};
		};

	/* --- forbid copy constructor --- */
	Poller( const Poller & )
	:	pPollerThread( 0 )
		{ assert( 0 ); };

public:
	Poller()
	:	pPollerThread( (UserThread *)Platform::AllocThread( 0, 0 ) )
		{
		PIMUSTBETRUE( pPollerThread );
		PIMUSTBETRUE( pPollerThread->Begin( 
			PollLoop,
			(unsigned long)this,
			PITHREAD_PRIORITY_LOWEST,
			PITHREAD_FLAGS_SUSPENDED ) );
		};

	~Poller()
		{
		PI_DELETE( pPollerThread );
		for( DblListIterator i( tPollingThreads ); !i.BadIndex(); i++ )
			{
			((UserThread *)i.Current())->MakeRunnable( -1 );
			};
		};

	void AddThread( UserThread *pThread )
		{
		assert( pThread && pThread->pPoll && pPollerThread );
		tPollingThreads.Append( (DblList::type)pThread );
		if ( pPollerThread->IsSuspended() )
			{ pPollerThread->Resume(); };
		};

	void RemoveThread( UserThread *pThread )
		{
		for( DblListIterator i( tPollingThreads ); !i.BadIndex(); i++ )
			{
			if ( pThread == i.Current() )
				{
				tPollingThreads.Detach( i );
				return;
				};
			};
		assert( 0 );
		};

	/* ---
	Loop once through the list of polling threads and issue a 
	select() to get first available ready descriptor if necessary
	Returning false from this function is considered serious and
	may cause the poller thread, or program to terminate.
	--- */
	bool DoPoll()
		{
		assert( tPollingThreads.Size() > 0 );
		int iFDMax = -1;
		timeval tv;
		tv.tv_sec = INT_MAX;
		tv.tv_usec = INT_MAX;
		timeval *pTv = 0;		/* --- initially set to infinite timeout --- */
		fd_set tRead, tWrite, tExcept;
		FD_ZERO( &tRead ); FD_ZERO( &tWrite ); FD_ZERO( &tExcept );
#if DEBUG>=20
		CERR << "Entering DoPoll(), tPollingThreads.Size() = "
			<< tPollingThreads.Size() << endl;
#endif
		DblListIterator i( tPollingThreads );
		for(; !i.BadIndex(); i++)
			{
			PollInfo *pPoll = ((UserThread *)i.Current())->pPoll;
			assert( pPoll );
			int iFD = pPoll->iFD;
			if ( pPoll->HasTimeout() &&
				pTv==0 )
				{
				/* --- assign timeout --- */
				pTv = &tv;
				};
			/* ---
			Check to see if this PollInfo has timed out. Don't
			bother if another PollInfo has already timed out
			--- */
			if (
				pTv &&
				tv.tv_sec && tv.tv_usec && 
				pPoll->HasTimeout() && !pPoll->MinTimeout( pTv ) )
				{
				/* ---
				technically, this poll has timed out, but that
				may have happened with recent processing so see what
				the status is after the select(). Set the timeout
				values to 0, so the select() will not wait.
				--- */
				tv.tv_sec = 0;
				tv.tv_usec = 0;
				};
			if ( iFD>=0 )
				{
				/* ---
				This is a poll on a real descriptor and not just a 
				sleep 'til timeout
				--- */
				FD_SET( iFD, &tExcept );
				if ( pPoll->iFlags & Platform::POLL_READ )
					{ FD_SET( iFD, &tRead ); }
				if ( pPoll->iFlags & Platform::POLL_WRITE )
					{ FD_SET( iFD, &tWrite ); }
				if ( iFDMax<iFD )
					{ iFDMax = iFD; };
				};
			};
#if DEBUG>=20
		CERR << "DoPoll before select()";
		if ( pTv )
			{
			CERR << " timeout = " << pTv->tv_sec << ".";
			CERR.width( 6 );
			CERR.fill( '0' );
			CERR << pTv->tv_usec;
			CERR.fill( ' ' );
			};
		CERR << endl;
#endif
		int iRet = 0;
		if ( iFDMax>=0 )
			{
			for(;;)
				{
				iRet = ::select( iFDMax+1, &tRead, &tWrite, &tExcept, pTv ); 
			
				if ( iRet==-1 && errno==EINTR )
					{
					/* ---
					An interrupt could well have changed the runnable status
					of another thread. Give the scheduler a chance to see
					before restarting the select.
					--- */
					StackContextSwitch();
					continue;
					};

				break;
				};
			};
#if DEBUG>=20
		CERR << "DoPoll after select(), iRet = " << iRet
			<< ", errno = " << errno << endl;
#endif
		switch( iRet )
			{
			case -1:	/* error */
				PIOSERR;/* but this is not a user thread */
				return false;

			case 0:		/* timeout */
			default:	/* some sockets/files are ready */
				;
			};
		/* ---
		Note that it is necessary to step through the list of threads
		without using an iterator. This is because the call to MakeRunnable()
		ultimately mutates the list thus invalidating the iterator
		--- */
		for( int j=0; j<tPollingThreads.Size(); j++ )
			{
			UserThread *pThread = (UserThread *)tPollingThreads[j];
			PollInfo *pPoll = pThread->pPoll;
			assert( pPoll );
			int iFD = pPoll->iFD;
			int iFlags = pPoll->iFlags;
			bool bTimedOut = pPoll->HasTimeout() && !pPoll->MinTimeout( 0 );
#if DEBUG>=20
			CERR << "Socket: " << iFD;
#endif
			if ( bTimedOut )
				{
#if DEBUG>=20
				CERR << ", TIMEOUT";
#endif
				/* --- 0==return code for timeout --- */
				pThread->MakeRunnable( 0 );
				j--;
				}
			else if ( iRet>0 && iFD>=0 )
				{
				int iPollResult = 0;
				if ( FD_ISSET( iFD, &tExcept ) )
					{
#if DEBUG>=20
					CERR << ", EXCEPTION";
#endif
					iPollResult = -1;
					}
				else
					{
					if ( ( iFlags & Platform::POLL_READ ) && 
						FD_ISSET( iFD, &tRead ) )
						{
#if DEBUG>=20
						CERR << ", READ";
#endif
						iPollResult = iPollResult | Platform::POLL_READ;
						};
					if ( ( iFlags & Platform::POLL_WRITE ) && 
						FD_ISSET( iFD, &tWrite ) )
						{
#if DEBUG>=20
						CERR << ", WRITE";
#endif
						iPollResult = iPollResult | Platform::POLL_WRITE;
						};
					};
				if ( iPollResult )
					{
					pThread->MakeRunnable( iPollResult );
					j--;
					};
				};
#if DEBUG>=20
			CERR << endl;
#endif
			};	/* --- loop over all sockets up to iFDMax --- */
		return true;
		};
};

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int PollInfo::Poll( UserThread *pThread, int iFD, int iFlags,
	int iTimeout )
{
	assert( ___pThePoller );
	return pThread->PutToSleep( iFD, iFlags, iTimeout );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void UserThread::Cleanup()
{
	assert( !bIsTerminated );

	/* --- cleanup on the death of this thread --- */
	bIsTerminated = true;

	/* --- notify any thread which is blocking on this thread --- */
	for(;;)
		{
		UserThread *pTheThread = tWaiting.SelectNextThread();
		if ( !pTheThread )
			{ break; };
		pTheThread->MakeRunnable( this );
		};

	/* --- abort any blocks --- */
	AbortBlocks( 0 );

	/* --- clear other values --- */
	ulData = 0;
#if DEBUG>=20
	CERR << "User Thread end of Cleanup(): " << (void *)this << endl;
	DbgWrite( CERR );
	CERR << endl;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	virtual, public
 Description:
\*___________________________________________________________________________*/
bool UserThread::SetPriority( int iThePriority )
{
	if (!(	iThePriority <= PITHREAD_PRIORITY_HIGHEST &&
			iThePriority >= PITHREAD_PRIORITY_LOW ) )
		{ return false; };
	int iOldPriority = iPriority; 
	InternalSetPriority( iThePriority );
	if ( ___pTheCurrentThread != this )
		{
		___pLatentThreads->ChangeThreadPriority( this, iOldPriority,
			iPriority );
		};

	/* --- assume correct priorities --- */
	StackContextSwitch();
	return true;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	private
 Description:
	Return true if the priority was changed, else false.
\*___________________________________________________________________________*/
bool UserThread::InternalSetPriority( int iNewPriority )
{
	if ( iNewPriority == iPriority )
		{ return false; };
#if DEBUG>=20
		CERR << (void *)this
			<< "UserThread::InternalSetPriority(): changing priority from "
			<< iPriority << " to " << iNewPriority << endl;
#endif
	/* ---
	Change the priority of this thread wherever else it is 
	recorded
	--- */
	if ( pThread )
		{
		pThread->tWaiting.ChangeThreadPriority( this, iPriority,
			iNewPriority );
		};
	if ( pSemaphore )
		{
		pSemaphore->tWaiting.ChangeThreadPriority( this, iPriority,
			iNewPriority );
		};
	iPriority = iNewPriority;
	return true;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
bool UserThread::InternalTerminate( unsigned long ulTheStatus )
{
	ulStatus = ulTheStatus;
	assert( !bIsTerminated );
	if ( ___pTheCurrentThread == this )
		{
		StackUnwind();
		return true;		/* this is never reached */
		}
	else
		{
		___pLatentThreads->RemoveThread( this );
		___pExitingThread = this;
		StackContextSwitch();
		return true;
		};
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
int UserThread::PutToSleep( int iFD, int iFlags, int iTimeout )
{
	pPoll = PI_NEW( PollInfo( iFD, iFlags, iTimeout ) );
	___pThePoller->AddThread( this );
	StackContextSwitch();
	int iTmp = iNextReturnCode;
	iNextReturnCode = 0;
	return iTmp;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void UserThread::MakeRunnable( int iThePollResult )
{
	assert( pPoll );
	___pThePoller->RemoveThread( this );
	PI_DELETE( pPoll );
	pPoll = 0;
	iNextReturnCode = iThePollResult;		
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void UserThread::AbortBlocks( int iWhy )
{
	/* --- notify any objects that we are blocking on --- */	

	/* --- thread --- */
	if ( pThread )
		{ pThread->GaveUpWaiting( this ); };
	pThread = 0;

	/* --- mutex and semaphores --- */
	if ( pSemaphore )
		{ pSemaphore->GaveUpWaiting( this ); };
	pSemaphore = 0;
	
	/* --- sleep, file descriptor poll --- */
	if ( pPoll )
		{ MakeRunnable( iWhy ); };

	/* --- suspension --- */
	bSuspended = false;

	iNextReturnCode = iWhy;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Get the register context block for a thread
\*___________________________________________________________________________*/
size_t *__Pi_GetThreadContext( void *pThread )
{
	return ((UserThread *)pThread )->GetContext();
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Add a thread into the pool of latent threads
\*___________________________________________________________________________*/
void __Pi_AddLatentThread( void *pThread )
{
	___pLatentThreads->AddThread( (UserThread *)pThread );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void __Pi_BeginThread( void *pThread )
{
	UserThread *pNewThread = (UserThread *)pThread;
	pNewThread->BeginThread();
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void __Pi_UnwindThreadStack( void *pThread )
{
	((UserThread *)pThread)->StackUnwind();
}
/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void *__Pi_SelectNextRunnableThread()
{
	/* ---
	Must get another thread or cause deadlock
	--- */
	return ___pLatentThreads->SelectNextRunnableThread( 1 );
}

/*___________________________________________________________________________*\
 *
 Function:	StackContextSwitch()
 Synopsis:
 Description:
	This is a very specical function. It swaps out the stack and so does
	a 'return' into another thread

	See Asm.cpp for an explaination.

	Returns non-zero if the thread was changed. This is useful internally
	for optimizing yield.

	NOTE: That optimization is disable for this entire file because of
	bad things that happen when ___pTheCurrent and ___pPlaceholer are
	optimized into local variables with __Pi_StackContextSwitch() in the
	middle.
\*___________________________________________________________________________*/
/* NOTE: temporary */
static UserThread *___pPlaceholder;
/* NOTE: end temporary */
int StackContextSwitch( void )
{
	int iFlag;

	UserThread *pNextThread = 0;
	if ( ___pNewThread )
		{
		iFlag = SCHED_BEGIN;
		pNextThread = ___pNewThread; 
		___pNewThread = 0;
		}
	else if ( ___pExitingThread )
		{
		iFlag = SCHED_EXIT;
		pNextThread = ___pExitingThread; 
		___pExitingThread = 0;
		}
	else
		{
		iFlag = SCHED_NEXT;
#if 0
		pNextThread = ___pLatentThreads->SelectNextRunnableThread( 0 );
		if ( !pNextThread )
			{
			/* --- don't switch context --- */
			return 0;
			};
#endif
		};

	___pPlaceholder = ___pTheCurrentThread;

#if DEBUG>=15
	CERR << "Before __Pi_StackContextSwitch()" << endl;
	DbgWriteAllThreads( CERR );
#endif

	__Pi_StackContextSwitch(
		(void **)&( ___pTheCurrentThread ),		/* address of thread pointer */
		pNextThread,	
		iFlag );								/* action to take */

#if DEBUG>=15
	CERR << "After __Pi_StackContextSwitch()" << endl;
	DbgWriteAllThreads( CERR );
#endif

#if 0
	return 1;
#else
	return ___pPlaceholder==___pTheCurrentThread ? 0 : 1;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void Internal_InitThreads()
{
	if ( ___pLatentThreads )
	assert( !___pLatentThreads );
	if ( ___pLatentThreads )
		{ return; /* already initialized */ };
	memset( aUsedKeys, 0, sizeof( int ) * PI_MAXTHREADKEYS );
	___pLatentThreads = PI_NEW( ThreadScheduler );

	/* ---
	create a structure for the current thread (the main thread)
	and set the global thread pointer 
	--- */
	___pTheMainThread = PI_NEW( UserThread(
					Platform::GetGlobalAllocator(), 0 ) );
	___pTheCurrentThread = ___pTheMainThread;
	___pThePoller = PI_NEW( Poller );
	assert( ___pThePoller );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	This function assumes that the main thread is executing this function.
\*___________________________________________________________________________*/
void Internal_CleanupThreads()
{
	assert( ___pLatentThreads );
	if ( !___pLatentThreads )
		{ return; /* never initialized */ };

	PI_DELETE( ___pThePoller );
	___pThePoller = 0;

	/* --- this must be the main thread --- */
	PIMUSTBETRUE( ___pTheMainThread == ___pTheCurrentThread );
	while( ___pLatentThreads->Size() )
		{ PI_DELETE( (UserThread *)___pLatentThreads->SelectNextThread() ); };	
	PI_DELETE( ___pLatentThreads );
	___pLatentThreads = 0;

	/* --- destroy the main thread object --- */
	PI_DELETE( ___pTheMainThread );
	___pTheMainThread = 0;
	___pTheCurrentThread = 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
void Platform::Internal_EnterThreads( void (* fn)(void) )
{
	/* ---
	Initialize global values, even tough there are declared static and
	so must be zero initialized on startup, the threads initialization	
	could happen multiple times, say because of hangup restart
	--- */
	___pLatentThreads = 0;			/* global thread scheduler */
	___pThePoller = 0;				/* polls on sockets/files */
	___pTheMainThread = 0;			/* the main thread */
	___pTheCurrentThread = 0;		/* the current thread */
	___pNewThread = 0;				/* a newly created thread */
	___pExitingThread = 0;			/* an exiting thread */
	iThreadsCreated = 0;			/* since initialization */

	Internal_InitThreads();
	if ( fn )
		{
		___pTheMainThread->Begin( (ThreadFn)fn,
			(unsigned long)0, PITHREAD_PRIORITY_MED, 0 );
		};
	Internal_CleanupThreads();
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
Allocator *Platform::GetCurrentAllocator()
{
	return Platform::GetGlobalAllocator();
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Semaphore *Platform::AllocLocalMutex()
{
	return PI_NEW( UserLocalSemaphore( 1, 1) );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Semaphore *Platform::AllocLocalSemaphore( int iInitialCount, int iMaxCount )
{
	return PI_NEW( UserLocalSemaphore( iInitialCount, iMaxCount ) );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Thread *Platform::AllocThread(
		void *pV,
		int iStackSize )
{
	Allocator *pAllocator = (Allocator *)pV;	/* LATER, for MSVC4.2 */
	Thread *pNewThread = PI_NEW( UserThread(
			pAllocator ? pAllocator : Platform::GetGlobalAllocator(),
			iStackSize ? iStackSize : PI_DEFAULTSTACKSIZE ) );
	return pNewThread;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::YieldThread()
{
	/* 
	** It would be good at some point to get a more sophisticated yield
	** with a flag to indicate different flavours
	**
	**
	** For Example
	** -----------
	** Add code to make this sleep( 0 ) to yield the process
	** timeslice if no threads are available.......Done
	**
	*/
	if ( !StackContextSwitch() )
		{
		/* ---
		Thread was not switched out, yield process timeslice
		--- */
#if defined(CONFIG_OS_POSIX)
#	if !defined(CONFIG_NO_YIELD)
		::yield();
#	endif
#else
#	if defined(CONFIG_OS_WIN32)
		::Sleep( 0 );
#	elif defined(CONFIG_OS_WIN16)
		::Yield();
#	endif
#endif
		};
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::WaitForThread( Thread *pThread )
{
	int iRet = ___pTheCurrentThread->WaitForThread( (UserThread *)pThread );
	if ( iRet!=0 )
		{ PIERROR(iRet); };
	return iRet;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Thread *Platform::GetCurrentThread()
{
	return ___pTheCurrentThread;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
Thread *Platform::GetMainThread()
{
	return ___pTheMainThread;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::AllocThreadKey()
{
	for( int i=0; i<PI_MAXTHREADKEYS; i++ )
		{
		if ( !aUsedKeys[i] )
			{
			aUsedKeys[i] = 1;
			return i;
			};
		};
	PIERROR( PIAPI_EXHAUSTED );
	return -1;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
bool Platform::DeleteThreadKey( int iKey )
{
	if ( iKey<0 || iKey>=PI_MAXTHREADKEYS || !aUsedKeys[iKey] )
		{
		PIERROR( PIAPI_EINVAL );
		return false;
		};
	aUsedKeys[iKey] = 0;
	return true;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::ThreadDbgDump( ostream &os )
{
	DbgWriteAllThreads( os );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::PollFD( int iFD, int iFlags, int iTimeout )
{
#if defined(CONFIG_OS_POSIX)
#	if defined(CONFIG_OS_SOLARIS)
		/*
		** Solaris 2.x: man -s 3C select
		**	
		** [File descriptors associated with regular files always select
		** true  for  ready  to  read, ready to write, and error conditions.]
		**
		** NOTE:
		**		- What about pipes?
		**		- Also change other instances of this function
		*/
		/* so just return the requested status */
		return iFlags;
#	endif
	return Platform::PollNetFD( iFD, iFlags, iTimeout );
#elif defined(CONFIG_OS_WIN32)
	/* ---
	Win32 SDK, says that WaitForSingleObject() can be used to
	check if a file handle operation will succeed without blocking
	'under certain circumstances' and is 'discouraged'. But there doesn't
	seem to be another way to do this without side effects.
	NOTE: The reason this is discouraged is because confusion can arise if
	multiple waits exists for the the same handle, therefore its
	suggested to use an event. Anyway.....
	--- */
	int iRetFlags = 0;
	switch( ::WaitForSingleObject( (HANDLE)iFD, iTimeout*1000 ) )
		{
		case WAIT_FAILED:
			PIOSERR;
			return -1;

		default:;
			/* ---
			Don't know what made the handle signalled so set the
			return code based on what the callee requested. Files
			are unidirectional anyway so this doesn't cause confusion
			like it would with sockets
			--- */
			if ( iFlags & Platform::POLL_READ )
				{ iRetFlags = iRetFlags | Platform::POLL_READ; };
			if ( iFlags & Platform::POLL_WRITE )
				{ iRetFlags = iRetFlags | Platform::POLL_WRITE; };
			return iRetFlags;
		};
#else
		/*
		** Return ready without yielding
		*/
		(void)iFD;
		(void)iTimeout;
		return iFlags;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::PollPipeFD( int iFD, int iFlags, int iTimeout )
{
#if defined(CONFIG_OS_WIN32)
	int iRetFlags = 0;

	if ( iFlags & ~Platform::POLL_WRITE )
		{
		int iRet = Platform::PollFD( iFD, iFlags & ~Platform::POLL_WRITE,
			iTimeout );

		if ( iRet<0 )
			{ return iRet; };

		iRetFlags |= iRet;
		};

	/* ---
	Named pipes on Win32. Always return 'Ready to Write'.
	--- */
	if ( iFlags & Platform::POLL_WRITE )
		{ iRetFlags |= Platform::POLL_WRITE; };

	return iRetFlags;
#endif
	/*
	** Otherwise behaviour is the same as Platform::PollFD
	*/
	return Platform::PollFD( iFD, iFlags, iTimeout );	
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::PollNetFD( int iFD, int iFlags, int iTimeout )
{
	if ( iFD<0 )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};
	int iRet = PollInfo::Poll( ___pTheCurrentThread, iFD, iFlags,
		(iTimeout==-1) ? -1 : (iTimeout * 1000) );
	if ( iRet<=0 )
		{ PIERROR(iRet); };
	return iRet;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::Sleep( int iMilliseconds )
{
	if ( iMilliseconds<0 )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};
	int iRet = PollInfo::Poll( ___pTheCurrentThread, -1, 0, iMilliseconds );
	if ( iRet!=0 )
		{ PIERROR(iRet); };
	return iRet;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::CatchExceptions( void (* fn)( void *), void *pArg )
{
	/* ---
	It is an internal error for this function to be used on 
	threads other than the main thread
	--- */
	assert( ___pTheCurrentThread==___pTheMainThread );
	assert( fn );
#if defined(CONFIG_CPP_EXCEPTIONS)
	try	{
		(fn)( pArg );
		}
	catch( UserThread::StackUnwindException *pE )
		{
		PI_DELETE( pE );
		}
	catch( ... )
		{
		};
#else
	if ( setjmp( ___pTheMainThread->tJmpBuf )==0 )
		{ (fn)( pArg ); };
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::Write( int iFd, int iLength, const void *pData )
{
	assert( iFd!=-1 && iLength && pData );
	if ( iFd==-1 || !iLength || !pData )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};
	int iToWrite = iLength;
	int iWritten;
	for(; iToWrite ;)
		{
		iWritten = -1;
		/* --- yield if necessary --- */
		if ( Platform::PollFD( iFd, POLL_WRITE, -1 )<=0 )
			{ break; };
#if defined(CONFIG_OS_POSIX)
		iWritten = ::write( iFd, pData, iLength );
		if ( iWritten==-1 && errno==EINTR )
			{ continue; };

#elif defined(CONFIG_OS_WIN32)
		/* --- 
		the cast from int * to DWORD * (DWORD is currrently unsigned long), is
		not	so bad because iLength is the maximum size that can be sent and 
		is also an int
		--- */
		if ( !::WriteFile( (HANDLE)iFd, pData, iLength, (DWORD *)&iWritten,
			0 ) )
			{ iWritten=-1; };

#endif
		if ( !iWritten || iWritten==-1 )
			{ break; };
		iToWrite -= iWritten;
		};

	if ( iWritten==-1 )
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::WriteAtomic( int iFd, int iLength, const void *pData )
{
	assert( iFd!=-1 && iLength && pData );
	if ( iFd==-1 || !iLength || !pData )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};
	int iToWrite = iLength;
	int iWritten;

	/* --- yield until write is writtable --- */
	if ( Platform::PollFD( iFd, POLL_WRITE, -1 )<=0 )
		{ assert( 0 ); return PIAPI_ERROR; };

	if ( Platform::LockFd( iFd )!=PIAPI_COMPLETED )
		{	
		return PIAPI_ERROR;
		};

	for(; iToWrite ;)
		{
		iWritten = -1;

#if defined(CONFIG_OS_POSIX)
		iWritten = ::write( iFd, pData, iLength );
		if ( iWritten==-1 && errno==EINTR )
			{ continue; };

#elif defined(CONFIG_OS_WIN32)
		/* --- 
		the cast from int * to DWORD * (DWORD is currrently unsigned long), is
		not	so bad because iLength is the maximum size that can be sent and 
		is also an int
		--- */
		if ( !::WriteFile( (HANDLE)iFd, pData, iLength, (DWORD *)&iWritten,
			0 ) )
			{ iWritten=-1; };

#endif
		if ( !iWritten || iWritten==-1 )
			{ break; };
		iToWrite -= iWritten;
		};

	iWritten = Platform::UnlockFd( iFd );
	if ( iWritten!=PIAPI_COMPLETED )
		{
		PIOSERR;
		return PIAPI_ERROR;
		};

	return PIAPI_COMPLETED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::WaitForProcess( int tProcess, int iFlags, int /* iTimeout */ )
{
	(void)iFlags;	/* not used */
#if defined(CONFIG_OS_POSIX)
	for(;;)
		{
		int iRet = ::waitpid( tProcess, 0, WNOHANG );
#if !defined(CONFIG_NO_ERESTART)
		if ( iRet==-1 && errno!=EINTR && errno!=ERESTART )
#else
		if ( iRet==-1 && errno!=EINTR )
#endif
			{
			PIOSERR;
			return PIAPI_ERROR;
			};

		if ( iRet==tProcess )
			{ return PIAPI_COMPLETED; };

		/* --- yield and try again --- */
		Platform::YieldThread();
		};

#elif defined(CONFIG_OS_WIN32)
	for(;;)
		{
		DWORD dwWait=::WaitForSingleObject( (HANDLE)tProcess, 0 /* no wait */ );
		switch( dwWait )
			{
			case WAIT_TIMEOUT:
				/* --- yield and try again --- */
				Platform::YieldThread();	// or yield timeslice
				break;

			case WAIT_OBJECT_0:
				return PIAPI_COMPLETED;

			case WAIT_FAILED:
				PIOSERR;
				return PIAPI_ERROR;

			default:
				/* how did we get here */
				assert( 0 );
				PIERROR( PIAPI_ERROR );
				return PIAPI_ERROR;
			};
		};
#endif
	(void)tProcess;
	return PIAPI_NOTSUPPORTED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
#if defined(CONFIG_LOCK_FLOCK)
#	define PI_LOCK_BLOCK		(LOCK_EX)
#	define PI_LOCK_NOBLOCK		(LOCK_EX | LOCK_NB)
#else
#	define PI_LOCK_BLOCK		(F_LOCK)
#	define PI_LOCK_NOBLOCK		(F_TLOCK)
#endif
int Platform::LockFd( int iFd )
{
	assert( iFd!=-1 );
	if ( iFd==-1 )
		{
		PIERROR( PIAPI_EINVAL );
		return -1;
		};

#if defined(CONFIG_OS_POSIX)
	/* ---
	Enter loop to obtain a lock on the file
	--- */
	int iFlags = PI_LOCK_NOBLOCK;
	for(;;)
		{
		/* --- non-blocking call to flock --- */
#if defined(CONFIG_LOCK_FLOCK)
		if ( ::flock( iFd, iFlags )==0 )
#else
		if ( ::lockf( iFd, iFlags, 0 )==0 )
#endif
			{
			/* --- got the lock, continue --- */
			break;
			}
		else if ( errno==EWOULDBLOCK || errno==EAGAIN )
			{
			/* ---
			Another process has the lock, try to yield the thread. 
			If the thread is not yieldm then block on the lock to 
			avoid a tight loop
			--- */
			if ( StackContextSwitch() )
				{
				/* --- thread yielded, don't block on retry --- */
				iFlags = PI_LOCK_NOBLOCK;
				}
			else
				{
				/* ----
				No other thread is runnable at this thread's priority
				level, so this is just a tight loop. Clear the extra
				flags to cause a block to occur
				--- */
				iFlags = PI_LOCK_BLOCK;
				};

			continue;
			}
		else if ( errno==EINTR )
			{
			/* ---
			man on solaris did not list this as a possible scenario,
			i.e. operation interrupted because a signal was caught,
			but just in case...
			--- */
			iFlags = PI_LOCK_NOBLOCK;
			continue;
			};

		/* --- an error occurred in locking --- */
		PIOSERR;
		return PIAPI_ERROR;
		};
#endif

	return PIAPI_COMPLETED;
}


/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::UnlockFd( int iFd )
{
	assert( iFd!=-1 );

#if defined(CONFIG_OS_POSIX)
	/* ---
	Release the lock
	--- */
	int iError = 0;
	for(;;)
		{
#if defined(CONFIG_LOCK_FLOCK)
		if ( ::flock( iFd, LOCK_UN )==0 )
#else
		if ( ::lockf( iFd, F_ULOCK, 0 )==0 )
#endif
			{ break; };
		
		if ( errno==EINTR )
			{ 
			Platform::YieldThread();
			continue;
			};

		/* --- real error */
		iError = -1;
		break;
		};

	if ( iError==-1 )
		{
		PIOSERR;
		assert( 0 );
		return PIAPI_ERROR;
		};
#endif

	return PIAPI_COMPLETED;
}

#endif /* defined(CONFIG_MULTITHREADED) && defined(CONFIG_MT_USER) */

